""" VOMSSecurityManager unit tests
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name,too-many-lines,no-value-for-parameter

import unittest
import stat

import mock
from DIRAC import S_OK, S_ERROR
import DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.VOMSSecurityManager

# This just defines a few groups with their VOMSRole
diracGrps = {
    "grp_admin": None,
    "grp_data": "vomsProd",
    "grp_mc": "vomsProd",
    "grp_user": "vomsUser",
    "grp_nothing": None,
}


# A dictionary of directories. The keys are path,
# the values are another dic with keys 'owner', 'OwnerGroup' and 'mode'
directoryTree = {}

# Same as directoryTree, but for files
fileTree = {}


def setupTree():
    """This method sets up the hierarchy, directories and files, and stores them in
    directoryTree and fileTree
    """

    def makeNode(owner, group, mode):
        """Just returns a dictionary, keys are the parm names, and associated to their values"""
        return {"owner": owner, "OwnerGroup": group, "mode": mode}

    def setupFiles():
        """Internal method adding the files"""

        global fileTree
        fileTree = {}

        fileTree["/atTheRoot.txt"] = makeNode("admin", "grp_admin", 0o775)

        fileTree["/realData/run1/run1_data.txt"] = makeNode("dm", "grp_data", 0o775)
        fileTree["/realData/run2/run2_data.txt"] = makeNode("dm", "grp_data", 0o705)

        # We should be able to replicate but not remove that file
        fileTree["/realData/run3/run3_data.txt"] = makeNode("dm", "grp_data", 0o775)

        fileTree["/users/usr1/usr1_file.txt"] = makeNode("usr1", "grp_user", 0o755)
        fileTree["/users/usr1/sub1/usr1_secret.txt"] = makeNode("usr1", "grp_user", 0o700)

        fileTree["/users/usr2/usr2_file.txt"] = makeNode("usr2", "grp_user", 0o700)

    global directoryTree
    directoryTree = {}

    # Only root and members from grp_admin should be able to create something in the root directory
    # Others cannot read it
    directoryTree["/"] = makeNode("admin", "grp_admin", 0o770)

    # groups with vomsProd should be able to create and remove directories
    # in /realData
    directoryTree["/realData"] = makeNode("dm", "grp_data", 0o775)
    directoryTree["/realData/run1"] = makeNode("dm", "grp_data", 0o775)
    directoryTree["/realData/run2"] = makeNode("dm", "grp_data", 0o775)
    # No one should be able to write in this one
    directoryTree["/realData/run3"] = makeNode("otherdm", "grp_data", 0o555)

    # Only root can create new user directories or remove them
    # Only the user can create or remove its subdirs
    directoryTree["/users"] = makeNode("admin", "grp_admin", 0o755)
    directoryTree["/users/usr1"] = makeNode("usr1", "grp_user", 0o755)
    directoryTree["/users/usr1/sub1"] = makeNode("usr1", "grp_user", 0o700)
    directoryTree["/users/usr2"] = makeNode("usr2", "grp_user", 0o755)
    directoryTree["/users/usr2/sub1"] = makeNode("usr2", "grp_user", 0o755)

    # grp_data and grp_mc atre both prodVoms role so should be able to write
    directoryTree["/mc"] = makeNode("mc1", "grp_mc", 0o775)
    directoryTree["/mc/prod1"] = makeNode("mc1", "grp_data", 0o775)
    directoryTree["/mc/prod2"] = makeNode("mc2", "grp_mc", 0o775)

    setupFiles()


# List of non existing directories.
nonExistingDirectories = ["/realData/futurRun", "/fakeBaseDir", "/users/usr1/subUsr1", "/users/usr2/subUsr2"]

# List of non existing files.
nonExistingFiles = ["/realData/futurRun/futur_data.txt", "/fakeBaseDir/fake_file.txt", "/fake_base.txt"]


class mock_DirectoryManager(object):
    """This class is a mock of a directory manager.
    It takes the information from directoryTree instead of the DB
    """

    def __init__(self):
        pass

    def exists(self, lfns):
        return S_OK({"Successful": dict((lfn, lfn in directoryTree) for lfn in lfns), "Failed": {}})

    def getDirectoryParameters(self, path):
        return S_OK(directoryTree[path]) if path in directoryTree else S_ERROR("Directory not found")

    def getDirectoryPermissions(self, path, credDict):
        if path not in directoryTree:
            return S_ERROR("Directory not found")

        owner = credDict["username"] == directoryTree[path]["owner"]
        group = credDict["group"] == directoryTree[path]["OwnerGroup"]
        mode = directoryTree[path]["mode"]
        resultDict = {}

        resultDict["Read"] = (
            (owner and mode & stat.S_IRUSR > 0) or (group and mode & stat.S_IRGRP > 0) or mode & stat.S_IROTH > 0
        )

        resultDict["Write"] = (
            (owner and mode & stat.S_IWUSR > 0) or (group and mode & stat.S_IWGRP > 0) or mode & stat.S_IWOTH > 0
        )

        resultDict["Execute"] = (
            (owner and mode & stat.S_IXUSR > 0) or (group and mode & stat.S_IXGRP > 0) or mode & stat.S_IXOTH > 0
        )

        return S_OK(resultDict)


class mock_FileManager(object):
    """This class is a mock of a file manager.
    It takes the information from fileTree instead of the DB
    """

    def __init__(self):
        pass

    def exists(self, lfns):
        return S_OK({"Successful": dict((lfn, lfn in fileTree) for lfn in lfns), "Failed": {}})

    def getFileMetadata(self, lfns):
        if not isinstance(lfns, list):
            lfns = [lfns]

        retParam = [
            "Size",
            "Checksum",
            "ChecksumType",
            "UID",
            "GID",
            "GUID",
            "CreationDate",
            "ModificationDate",
            "Mode",
            "Status",
        ]

        successful = {}
        failed = {}
        for filename in lfns:
            if filename not in fileTree:
                failed[filename] = "No such file or directory"
                continue

            val = dict.fromkeys(retParam, "mockValues")
            val["Owner"] = fileTree[filename]["owner"]
            val["OwnerGroup"] = fileTree[filename]["OwnerGroup"]
            val["Mode"] = fileTree[filename]["mode"]

            successful[filename] = val

        return S_OK({"Successful": successful, "Failed": failed})

    def getPathPermissions(self, lfns, credDict):
        if not isinstance(lfns, list):
            lfns = [lfns]

        successful = {}
        failed = {}
        for filename in lfns:

            if filename not in fileTree:
                failed[filename] = "File not found"
                continue

            owner = credDict["username"] == fileTree[filename]["owner"]
            group = credDict["group"] == fileTree[filename]["OwnerGroup"]
            mode = fileTree[filename]["mode"]

            resultDict = {}

            resultDict["Read"] = (
                (owner and mode & stat.S_IRUSR > 0) or (group and mode & stat.S_IRGRP > 0) or mode & stat.S_IROTH > 0
            )

            resultDict["Write"] = (
                (owner and mode & stat.S_IWUSR > 0) or (group and mode & stat.S_IWGRP > 0) or mode & stat.S_IWOTH > 0
            )

            resultDict["Execute"] = (
                (owner and mode & stat.S_IXUSR > 0) or (group and mode & stat.S_IXGRP > 0) or mode & stat.S_IXOTH > 0
            )

            successful[filename] = resultDict

        return S_OK({"Successful": successful, "Failed": failed})


class mock_db(object):
    """This class is a mock of a FileCatalogDB.
    It just contains dtree and fileManager references
    and sets the globalReadAccess to False
    """

    def __init__(self):
        self.globalReadAccess = False
        self.dtree = mock_DirectoryManager()
        self.fileManager = mock_FileManager()


class mock_SecurityManagerBase(object):
    """This class is a mock of a security manager.
    It just mockes the hasAdminAccess method
    """

    def __init__(self, database=False):
        self.db = mock_db()

    def hasAdminAccess(self, credDict):
        """Returns true only if the group is grp_admin"""
        return S_OK(credDict["group"] == "grp_admin")


# The mock module could not trick the inheritance mechanism, so I do it myself
DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.VOMSSecurityManager.VOMSSecurityManager.__bases__ = (
    mock_SecurityManagerBase,
)


def mock_getAllGroups():
    """Mocks the getAllGroups method from the CS"""
    return list(diracGrps)


def mock_getGroupOption(grpName, grpOption):
    """Mocks the getGroupOption method from the CS,
    only for vomsRole option
    """
    return diracGrps[grpName]


class BaseCaseMixin(object):
    """Base test class. Defines all the method to test"""

    @mock.patch(
        "DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.VOMSSecurityManager.getGroupOption",
        side_effect=mock_getGroupOption,
    )
    @mock.patch(
        "DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.VOMSSecurityManager.getAllGroups",
        side_effect=mock_getAllGroups,
    )
    def setUp(self, _a, _b):

        global directoryTree
        global fileTree
        # A dictionary of directories. The keys are path,
        # the values are another dic with keys 'owner', 'OwnerGroup' and 'mode'
        directoryTree = {}

        # Same as directoryTree, but for files
        fileTree = {}

        setupTree()

        # These two dict have to be defined by the children class and method,
        # and are used to compare against the output of the test
        self.expectedExistingRet = None
        self.expectedNonExistingRet = None

        # Manager object
        self.securityManager = (
            DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.VOMSSecurityManager.VOMSSecurityManager()
        )
        self.credDict = {"username": "anon", "group": "grp_admin"}

    def callForDirectories(self, methodName):
        """Test the permissions of a given file catalog method
        against all the existing and non existing directories
        """

        self.existingRet = self.securityManager.hasAccess(methodName, list(directoryTree), self.credDict)
        self.nonExistingRet = self.securityManager.hasAccess(methodName, nonExistingDirectories, self.credDict)

    def callForFiles(self, methodName):
        """Test the permissions of a given file catalog method
        against all the existing and non existing files
        """

        self.existingRet = self.securityManager.hasAccess(methodName, list(fileTree), self.credDict)
        self.nonExistingRet = self.securityManager.hasAccess(methodName, nonExistingFiles, self.credDict)

    def compareResult(self):
        """Compares the result between the test output and the expected values"""

        for testSet, real, expected in [
            ("Existing", self.existingRet, self.expectedExistingRet),
            ("NonExisting", self.nonExistingRet, self.expectedNonExistingRet),
        ]:

            self.assertTrue(real, "The method was not run")
            self.assertTrue(expected, "No expected results given")

            self.assertTrue(real["OK"] == expected["OK"], real)

            for dic in ["Successful", "Failed"]:
                dicReal = real["Value"][dic]
                dicExpected = expected["Value"][dic]

                notExpected = set(dicReal) - set(dicExpected)
                self.assertTrue(
                    notExpected == set(), "(%s) Returned more keys in %s than expected %s" % (testSet, dic, notExpected)
                )

                notReturned = set(dicExpected) - set(dicReal)
                self.assertTrue(notReturned == set(), "Some keys in %s are missing %s" % (dic, notReturned))

                for k in dicReal:
                    self.assertTrue(
                        dicReal[k] == dicExpected[k],
                        "(%s) Incompatible result for %s:\n\treal : %s\n\texpected %s"
                        % (testSet, k, dicReal[k], dicExpected[k]),
                    )

    def test_removeDirectory(self):
        """This is to test the special policy for removing directories"""

        self.callForDirectories("removeDirectory")
        self.compareResult()

    def test_createDirectory(self):
        """This is to test the default write policy for directories"""

        self.callForDirectories("createDirectory")
        self.compareResult()

    def test_listDirectory(self):
        """This is to test the special policy for listing directories"""

        self.callForDirectories("listDirectory")
        self.compareResult()

    def test_getDirectorySize(self):
        """This is to test the default read policy for directories"""

        self.callForDirectories("getDirectorySize")
        self.compareResult()

    def test_addFile(self):
        """This is to test the default write policy for files"""

        self.callForFiles("addFile")
        self.compareResult()

    def test_removeFile(self):
        """This is to test the special policy for removing files"""

        self.callForFiles("removeFile")
        self.compareResult()

    def test_getFileSize(self):
        """This is to test the default read policy for files"""

        self.callForFiles("getFileSize")
        self.compareResult()

    def test_changePathOwner(self):
        """This is to test policy for which we need to be admin"""

        self.callForFiles("changePathOwner")
        self.compareResult()

    def test_getReplicas(self):
        """This is to test the default read policy for replicas"""

        self.callForFiles("getReplicas")
        self.compareResult()

    def test_addReplica(self):
        """This is to test the default write policy for replicas"""

        self.callForFiles("addReplica")
        self.compareResult()


class TestNonExistingUser(BaseCaseMixin, unittest.TestCase):
    """As anonymous user and no group"""

    def setUp(self):
        super(TestNonExistingUser, self).setUp()
        self.credDict = {"username": "anon", "group": "grp_nothing"}

    def test_removeDirectory(self):
        """Removing directory with (anon, grp_nothing)
        Anonymous should not be able to remove any existing directory, but has
        the permission to remove non existing one
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestNonExistingUser, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (anon, grp_nothing)
        An anonymous user should not be able to create anything
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, False), "Failed": {}})

        super(TestNonExistingUser, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (anon, grp_nothing)
        Anonymous should not be granted the other bit
        """
        existingDic = {
            "/": False,
            "/realData": True,
            # owner of /realData is dm
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": True,
            "/users/usr1": True,
            "/users/usr1/sub1": False,  # sub1 is 700
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            # Owner group of /mc has vomsProd as well
            "/mc": True,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestNonExistingUser, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (anon, grp_nothing)
        Anonymous should not be granted the other bit
        """
        existingDic = {
            "/": False,
            "/realData": False,  # / is not readable by other
            # owner of /realData is dm
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,  # / is not readable by other
            "/users/usr1": True,
            "/users/usr1/sub1": True,  # usr1 is readable by all
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            # Owner group of /mc has vomsProd as well
            "/mc": False,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestNonExistingUser, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (anon, grp_nothing)
        An anonymous user should not be able to create anything
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, False), "Failed": {}})

        super(TestNonExistingUser, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (anon, grp_nothing)
        An anonymous user should not be able to remove anything,
        except non existing files
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestNonExistingUser, self).test_removeFile()

    def test_getFileSize(self):
        """Checking file size with (anon, grp_nothing)"""
        existingDic = {
            "/atTheRoot.txt": False,  # No read permission on /
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": False,  # sub1 is 700
            "/users/usr2/usr2_file.txt": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,  # Read permission on /realdata
            "/fakeBaseDir/fake_file.txt": False,  # No read permission on /
            "/fake_base.txt": False,  # No read permission on /
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestNonExistingUser, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (anon, grp_nothing)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, False), "Failed": {}})

        super(TestNonExistingUser, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (anon, grp_nothing)"""
        existingDic = {
            "/atTheRoot.txt": True,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": False,  # usr1_secret.txt is 700
            "/users/usr2/usr2_file.txt": False,  # usr2_file.txt is 700
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestNonExistingUser, self).test_getReplicas()

    def test_addReplica(self):
        """Adding replicas with (anon, grp_nothing)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": False,
            "/realData/run2/run2_data.txt": False,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": False,
            "/users/usr1/sub1/usr1_secret.txt": False,  # usr1_secret.txt is 700
            "/users/usr2/usr2_file.txt": False,  # usr2_file.txt is 700
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestNonExistingUser, self).test_addReplica()


class TestAdminGrpAnonUser(BaseCaseMixin, unittest.TestCase):
    """The grp_admin has adminAccess so should be able to do everything"""

    def setUp(self):
        super(TestAdminGrpAnonUser, self).setUp()
        self.credDict = {"username": "anon", "group": "grp_admin"}

    def test_removeDirectory(self):
        """Removing directory with (anon, grp_admin)
        The grp_admin has adminAccess so should be able to do everything
        """

        existingDic = dict.fromkeys(directoryTree, True)

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (anon, grp_admin)
        Everything should be granted
        """

        existingDic = dict.fromkeys(directoryTree, True)

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (anon, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (anon, grp_admin)
        Everything should be granted
        """
        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (anon, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (anon, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_removeFile()

    def test_getFileSize(self):
        """Checking file size with (anon, grp_admin)"""
        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (anon, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (anon, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_getReplicas()

    def test_addReplica(self):
        """Adding  replicas with (anon, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAnonUser, self).test_addReplica()


class TestAdminGrpAdminUser(BaseCaseMixin, unittest.TestCase):
    """The grp_admin has adminAccess so should be able to do everything"""

    def setUp(self):
        super(TestAdminGrpAdminUser, self).setUp()
        self.credDict = {"username": "admin", "group": "grp_admin"}

    def test_removeDirectory(self):
        """Removing directory with (admin, grp_admin)
        Everything should be granted
        """

        existingDic = dict.fromkeys(directoryTree, True)

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (admin, grp_admin)
        Everything should be granted
        """

        existingDic = dict.fromkeys(directoryTree, True)

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (admin, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (admin, grp_admin)
        Everything should be granted
        """
        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(directoryTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (admin, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (admin, grp_admin)
        Everything should be granted
        """

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_removeFile()

    def test_getFileSize(self):
        """Adding files with (admin, grp_admin)"""
        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (admin, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (admin, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_getReplicas()

    def test_addReplica(self):
        """Adding replicas with (admin, grp_admin)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, True), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestAdminGrpAdminUser, self).test_addReplica()


class TestDataGrpDmUser(BaseCaseMixin, unittest.TestCase):
    """Should have the permission of the 'dm' user and the group
    permission of all the vomsRole 'prod' (grp_data, grp_mc)
    """

    def setUp(self):
        super(TestDataGrpDmUser, self).setUp()
        self.credDict = {"username": "dm", "group": "grp_data"}

    def test_removeDirectory(self):
        """Removing directory with (dm, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            # owner of /realData is dm
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": False,
            "/users/usr1/sub1": False,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            # Owner group of /mc has vomsProd as well
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestDataGrpDmUser, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (dm, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            # owner of /realData is dm
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": False,
            "/users/usr1/sub1": False,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            # Owner group of /mc has vomsProd as well
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": False,
            "/users/usr2/subUsr2": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (dm, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": True,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": True,
            "/users/usr1": True,
            "/users/usr1/sub1": False,  # sub1 is 700
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": True,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (dm, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": True,
            "/users/usr1/sub1": True,
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": False,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (dm, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": False,
            "/users/usr1/sub1/usr1_secret.txt": False,  # sub1 is 700
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (dm, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": False,
            "/users/usr1/sub1/usr1_secret.txt": False,  # sub1 is 700
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestDataGrpDmUser, self).test_removeFile()

    def test_getFileSize(self):
        """Checking file size with (dm, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": False,  # sub1 is 700
            "/users/usr2/usr2_file.txt": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (dm, grp_data)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, False), "Failed": {}})

        super(TestDataGrpDmUser, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (dm, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": True,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": False,  # usr1_secret.txt is 700
            "/users/usr2/usr2_file.txt": False,  # usr2_file.txt is 700
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_getReplicas()

    def test_addReplica(self):
        """Adding replicas with (dm, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": False,
            "/users/usr1/sub1/usr1_secret.txt": False,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpDmUser, self).test_addReplica()


class TestDataGrpUsr1User(BaseCaseMixin, unittest.TestCase):
    """Should have the permission of the 'usr1' user and the group
    permission of all the vomsRole 'prod' (grp_data, grp_mc)
    """

    def setUp(self):
        super(TestDataGrpUsr1User, self).setUp()
        self.credDict = {"username": "usr1", "group": "grp_data"}

    def test_removeDirectory(self):
        """Removing directory with (usr1, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": False,
            # usr1 is the owner of /users/usr1
            "/users/usr1/sub1": True,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            # Owner group of /mc has vomsProd as well
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestDataGrpUsr1User, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (usr1, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            # owner of /realData is dm
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": False,
            "/users/usr1/sub1": True,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            # Owner group of /mc has vomsProd as well
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,  # Owner group of /realData is grp_data
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,  # Owner of /users/usr1 is usr1
            "/users/usr2/subUsr2": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (usr1, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": True,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": True,
            "/users/usr1": True,
            "/users/usr1/sub1": True,
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": True,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (usr1, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": True,
            "/users/usr1/sub1": True,
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": False,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (usr1, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (usr1, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestDataGrpUsr1User, self).test_removeFile()

    def test_getFileSize(self):
        """Checking file size with (usr1, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,  # Because other has it...
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (usr1, grp_data)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, False), "Failed": {}})

        super(TestDataGrpUsr1User, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (usr1, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": True,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,  # usr1_secret.txt is 700
            "/users/usr2/usr2_file.txt": False,  # usr2_file.txt is 700
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_getReplicas()

    def test_addReplica(self):
        """Adding replicas with (usr1, grp_data)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": False,  # Group has 0
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestDataGrpUsr1User, self).test_addReplica()


class TestUserGrpUsr1User(BaseCaseMixin, unittest.TestCase):
    """Just a normal user, should be able to write only in its own directory"""

    def setUp(self):
        super(TestUserGrpUsr1User, self).setUp()
        self.credDict = {"username": "usr1", "group": "grp_user"}

    def test_removeDirectory(self):
        """Removing directory with (usr1, grp_user)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": False,
            "/realData/run2": False,
            "/realData/run3": False,
            "/users": False,
            "/users/usr1": False,
            "/users/usr1/sub1": True,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            "/mc/prod1": False,
            "/mc/prod2": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingDirectories, True), "Failed": {}})

        super(TestUserGrpUsr1User, self).test_removeDirectory()

    def test_createDirectory(self):
        """Creating directory with (usr1, grp_user)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": False,
            "/realData/run2": False,
            "/realData/run3": False,
            "/users": False,
            "/users/usr1": False,
            "/users/usr1/sub1": True,
            "/users/usr2": False,
            "/users/usr2/sub1": False,
            "/mc": False,
            "/mc/prod1": False,
            "/mc/prod2": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": False,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_createDirectory()

    def test_listDirectory(self):
        """Listing directory with (usr1, grp_user)"""

        existingDic = {
            "/": False,
            "/realData": True,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": True,
            "/users/usr1": True,
            "/users/usr1/sub1": True,
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": True,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_listDirectory()

    def test_getDirectorySize(self):
        """Getting directory size with (usr1, grp_data)"""

        existingDic = {
            "/": False,
            "/realData": False,
            "/realData/run1": True,
            "/realData/run2": True,
            "/realData/run3": True,
            "/users": False,
            "/users/usr1": True,
            "/users/usr1/sub1": True,
            "/users/usr2": True,
            "/users/usr2/sub1": True,
            "/mc": False,
            "/mc/prod1": True,
            "/mc/prod2": True,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun": True,
            "/fakeBaseDir": False,
            "/users/usr1/subUsr1": True,
            "/users/usr2/subUsr2": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_getDirectorySize()

    def test_addFile(self):
        """Adding files with (usr1, grp_user)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": False,
            "/realData/run2/run2_data.txt": False,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": False,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_addFile()

    def test_removeFile(self):
        """Removing files with (usr1, grp_user)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": False,
            "/realData/run2/run2_data.txt": False,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, True), "Failed": {}})

        super(TestUserGrpUsr1User, self).test_removeFile()

    def test_getFileSize(self):
        """Checking file size with (usr1, grp_user)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": True,  # usr2 is readable by the group
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": False,
            "/fake_base.txt": False,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_getFileSize()

    def test_changePathOwner(self):
        """Setting fiel owner with (usr1, grp_user)"""

        self.expectedExistingRet = S_OK({"Successful": dict.fromkeys(fileTree, False), "Failed": {}})

        self.expectedNonExistingRet = S_OK({"Successful": dict.fromkeys(nonExistingFiles, False), "Failed": {}})

        super(TestUserGrpUsr1User, self).test_changePathOwner()

    def test_getReplicas(self):
        """Checking get replicas with (usr1, grp_user)"""
        existingDic = {
            "/atTheRoot.txt": True,
            "/realData/run1/run1_data.txt": True,
            "/realData/run2/run2_data.txt": True,
            "/realData/run3/run3_data.txt": True,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,  # usr1_secret.txt is 700
            "/users/usr2/usr2_file.txt": False,  # usr2_file.txt is 700
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_getReplicas()

    def test_addReplica(self):
        """Adding replicas with (usr1, grp_user)"""
        existingDic = {
            "/atTheRoot.txt": False,
            "/realData/run1/run1_data.txt": False,
            "/realData/run2/run2_data.txt": False,
            "/realData/run3/run3_data.txt": False,
            "/users/usr1/usr1_file.txt": True,
            "/users/usr1/sub1/usr1_secret.txt": True,
            "/users/usr2/usr2_file.txt": False,
        }

        self.expectedExistingRet = S_OK({"Successful": existingDic, "Failed": {}})

        nonExistingDic = {
            "/realData/futurRun/futur_data.txt": True,
            "/fakeBaseDir/fake_file.txt": True,
            "/fake_base.txt": True,
        }

        self.expectedNonExistingRet = S_OK({"Successful": nonExistingDic, "Failed": {}})

        super(TestUserGrpUsr1User, self).test_addReplica()


if __name__ == "__main__":

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestNonExistingUser)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestAdminGrpAnonUser))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestAdminGrpAdminUser))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestDataGrpDmUser))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestDataGrpUsr1User))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestUserGrpUsr1User))

    unittest.TextTestRunner(verbosity=2).run(suite)



""" This is a test of the parametric job generation tools
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest

from DIRAC.WorkloadManagementSystem.Utilities.ParametricJob import generateParametricJobs, getParameterVectorLength
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd

TEST_JDL_NO_PARAMETERS = """
[
  Executable = "my_executable";
  Arguments = "%s";
  JobName = "Test_%n";
]
"""

TEST_JDL_SIMPLE = """
[
  Executable = "my_executable";
  Arguments = "%s";
  JobName = "Test_%n";
  Parameters = { "a", "b", "c" }
]
"""

TEST_JDL_SIMPLE_BUNCH = """
[
  Executable = "my_executable";
  Arguments = "%s";
  JobName = "Test_%n";
  Parameters = 3;
  ParameterStart = 5;
]
"""

TEST_JDL_SIMPLE_PROGRESSION = """
[
  Executable = "my_executable";
  Arguments = "%s";
  JobName = "Test_%n";
  Parameters = 3;
  ParameterStart = 1;
  ParameterStep = 1;
  ParameterFactor = 2;
]
"""

TEST_JDL_MULTI = """
[
  Executable = "my_executable";
  Arguments = "%(A)s %(B)s";
  JobName = "Test_%n";
  Parameters = 3;
  ParameterStart.A = 1;
  ParameterStep.A = 1;
  ParameterFactor.A = 2;
  Parameters.B = { "a","b","c" };
]
"""

TEST_JDL_MULTI_BAD = """
[
  Executable = "my_executable";
  Arguments = "%(A)s %(B)s";
  JobName = "Test_%n";
  Parameters = 3;
  ParameterStart.A = 1;
  ParameterStep.A = 1;
  ParameterFactor.A = 2;
  Parameters.B = { "a","b","c","d" };
]
"""


class TestParametricUtilityCase(unittest.TestCase):
    def test_Simple(self):

        clad = ClassAd(TEST_JDL_SIMPLE)
        result = getParameterVectorLength(clad)
        self.assertTrue(result["OK"])
        nParam = result["Value"]

        self.assertEqual(nParam, 3)

        result = generateParametricJobs(clad)
        self.assertTrue(result["OK"])

        jobDescList = result["Value"]
        self.assertEqual(nParam, len(jobDescList))

        # Check the definition of the 2nd job
        jobClassAd = ClassAd(jobDescList[1])
        self.assertEqual(jobClassAd.getAttributeString("Arguments"), "b")
        self.assertEqual(jobClassAd.getAttributeString("JobName"), "Test_1")

    def test_SimpleBunch(self):

        clad = ClassAd(TEST_JDL_SIMPLE_BUNCH)
        result = getParameterVectorLength(clad)
        self.assertTrue(result["OK"])
        nParam = result["Value"]

        self.assertEqual(nParam, 3)

        result = generateParametricJobs(clad)
        self.assertTrue(result["OK"])

        jobDescList = result["Value"]
        self.assertEqual(nParam, len(jobDescList))

        # Check the definition of the 2nd job
        jobClassAd = ClassAd(jobDescList[1])
        self.assertEqual(jobClassAd.getAttributeString("Arguments"), "5")
        self.assertEqual(jobClassAd.getAttributeString("JobName"), "Test_1")

    def test_SimpleProgression(self):

        clad = ClassAd(TEST_JDL_SIMPLE_PROGRESSION)
        result = getParameterVectorLength(clad)
        self.assertTrue(result["OK"])
        nParam = result["Value"]

        self.assertEqual(nParam, 3)

        result = generateParametricJobs(clad)
        self.assertTrue(result["OK"])

        jobDescList = result["Value"]
        self.assertEqual(nParam, len(jobDescList))

        # Check the definition of the 2nd job
        jobClassAd = ClassAd(jobDescList[1])
        self.assertEqual(jobClassAd.getAttributeString("Arguments"), "3")
        self.assertEqual(jobClassAd.getAttributeString("JobName"), "Test_1")

    def test_Multi(self):

        clad = ClassAd(TEST_JDL_MULTI)
        result = getParameterVectorLength(clad)
        self.assertTrue(result["OK"])
        nParam = result["Value"]

        self.assertEqual(nParam, 3)

        result = generateParametricJobs(clad)
        self.assertTrue(result["OK"])

        jobDescList = result["Value"]
        self.assertEqual(nParam, len(jobDescList))

        # Check the definition of the 2nd job
        jobClassAd = ClassAd(jobDescList[1])
        self.assertEqual(jobClassAd.getAttributeString("Arguments"), "3 b")
        self.assertEqual(jobClassAd.getAttributeString("JobName"), "Test_1")

    def test_MultiBad(self):

        clad = ClassAd(TEST_JDL_MULTI_BAD)
        result = getParameterVectorLength(clad)
        self.assertTrue(not result["OK"])

    def test_NoParameters(self):

        clad = ClassAd(TEST_JDL_NO_PARAMETERS)
        result = getParameterVectorLength(clad)
        self.assertTrue(result["OK"])
        nParam = result["Value"]
        self.assertTrue(nParam is None)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestParametricUtilityCase)
    # suite.addTest( unittest.defaultTestLoader.loadTestsFromTestCase( xxxx ) )

    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



"""Module level filter."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels

DOT = "."
LEVEL = "__level__"


class ModuleFilter(object):
    """Filter module to set loglevel per module.

    ::

      Resources
      {
        LogBackends
        {
          <backend>
          {
             Filter = MyModuleFilter
          }
        }
        LogFilters
        {
           MyModuleFilter
           {
              Plugin = ModuleFilter
              dirac = ERROR
              dirac.Subprocess = DEBUG
              dirac.ILCDIRAC.Interfaces.API.NewInterface = INFO
           }
        }
      }

    This results in all debug messages from the Subprocess module to be printed, but only errors from
    the rest of dirac.  And INFO from a module in an extension. For this to work the global log level
    needs to be DEBUG (e.g., -ddd for commands)

    """

    def __init__(self, optionDict):
        """Contruct the object, set the base LogLevel to DEBUG, and parse the options."""
        self._configDict = {"dirac": {LEVEL: LogLevels.DEBUG}}
        optionDict.pop("Plugin", None)
        for module, level in optionDict.items():
            self.__fillConfig(self._configDict, module.split(DOT), LogLevels.getLevelValue(level))

    def __fillConfig(self, baseDict, modules, level):
        """Fill the config Dict with the module information.

        Recursivly fill the dictionary for each submodule with given level.
        If intermediate modules are not set, use DEBUG

        :param dict baseDict: dictionary for current submodules
        :param list modules: list of submodule paths to be set
        :parma int levelno: level to be set for given module
        """
        if len(modules) == 1:  # at the end for this setting
            if modules[0] in baseDict:
                baseDict[modules[0]][LEVEL] = level
            else:
                baseDict[modules[0]] = {LEVEL: level}
            return None
        module0 = modules[0]
        modules = modules[1:]
        if module0 not in baseDict:
            # DEBUG is the default loglevel for the root logger
            baseDict[module0] = {LEVEL: LogLevels.DEBUG}
        return self.__fillConfig(baseDict[module0], modules, level)

    def __filter(self, baseDict, hierarchy, levelno):
        """Check if sublevels are defined, or return highest set level.

        Recursively go through the configured levels, returns comparison with deepest match

        :param dict baseDict: dictionary with information starting at current level
        :param list hierarchy: list of module hierarchy
        :param int levelno: integer log level of given record
        :returns: boolean for filter value
        """
        if not hierarchy:
            return baseDict.get(LEVEL, -1) <= levelno
        if hierarchy[0] in baseDict:
            return self.__filter(baseDict[hierarchy[0]], hierarchy[1:], levelno)
        return baseDict.get(LEVEL, -1) <= levelno

    def filter(self, record):
        """Filter records based on the path of the logger."""
        return self.__filter(self._configDict, record.name.split(DOT), record.levelno)



"""
This module contains constants and lists for the possible job minor statuses.
"""
__RCSID__ = "$Id$"

#
APPLICATION = "Application"
#
APP_NOT_FOUND = "Application not found"
#
APP_THREAD_FAILED = "Application thread failed"
#
APP_THREAD_NOT_COMPLETE = "Application thread did not complete"
#
APP_SUCCESS = "Application Finished Successfully"
#
APP_ERRORS = "Application Finished With Errors"
#
DOWNLOADING_INPUT_SANDBOX = "Downloading InputSandbox"
#
DOWNLOADING_INPUT_SANDBOX_LFN = "Downloading InputSandbox LFN(s)"
#
FAILED_DOWNLOADING_INPUT_SANDBOX = "Failed Downloading InputSandbox"
#
FAILED_DOWNLOADING_INPUT_SANDBOX_LFN = "Failed Downloading InputSandbox LFN(s)"
#
FAILED_SENDING_REQUESTS = "Failed sending requests"
#
EXEC_COMPLETE = "Execution Complete"
#
EXCEPTION_DURING_EXEC = "Exception During Execution"
#
GOING_RESCHEDULE = "Going to reschedule job"
#
INPUT_DATA_RESOLUTION = "Input Data Resolution"
#
PILOT_AGENT_SUBMISSION = "Pilot Agent Submission"
#
PENDING_REQUESTS = "Pending Requests"
#
JOB_EXCEEDED_CPU = "Job has reached the CPU limit of the queue"
#
JOB_EXCEEDED_WALL_CLOCK = "Job has exceeded maximum wall clock time"
#
JOB_INITIALIZATION = "Job Initialization"
#
JOB_INSUFFICIENT_DISK = "Job has insufficient disk space to continue"
#
JOB_WRAPPER_INITIALIZATION = "Job Wrapper Initialization"
#
JOB_WRAPPER_EXECUTION = "JobWrapper execution"
#
MARKED_FOR_TERMINATION = "Marked for termination"
#
RECEIVED_KILL_SIGNAL = "Received Kill signal"
#
RESOLVING_OUTPUT_SANDBOX = "Resolving Output Sandbox"
#
STALLED_PILOT_NOT_RUNNING = "Job stalled: pilot not running"
#
UPLOADING_OUTPUT_SANDBOX = "Uploading Output Sandbox"
#
OUTPUT_SANDBOX_UPLOADED = "Output Sandbox Uploaded"
#
UPLOADING_OUTPUT_DATA = "Uploading Output Data"
#
UPLOADING_JOB_OUTPUTS = "Uploading Outputs"
#
OUTPUT_DATA_UPLOADED = "Output Data Uploaded"
#
WATCHDOG_STALLED = "Watchdog identified this job as stalled"
#
ILLEGAL_JOB_JDL = "Illegal Job JDL"
#
INPUT_NOT_AVAILABLE = "Input Data Not Available"
#
INPUT_CONTAINS_SLASHES = "Input data contains //"
#
INPUT_INCORRECT = "Input data not correctly specified"
#
NO_CANDIDATE_SITE_FOUND = "No candidate sites available"



"""Logging Filter based on pattern."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


class PatternFilter(object):
    """Filter module to set loglevel per module.

    ::

      Resources
      {
        LogBackends
        {
          <backend>
          {
            Filter = MyPatternFilter
          }
        }
        LogFilters
        {
           MyPatternFilter
           {
             Plugin = PatternFilter
             Accept = Some, Words, To, Match
             Reject = Foo, Bar, Baz, DEBUG
           }
        }
      }

    Only print log lines matching words in Accept and reject lines which contains words in Reject.
    A line must both be accepted and not rejected to be passed!
    """

    def __init__(self, optionDict):
        """Contruct the object, set the base LogLevel to DEBUG, and parse the options."""

        self._accept = [w.strip() for w in optionDict.get("Accept", "").split(",") if w.strip()]
        self._reject = [w.strip() for w in optionDict.get("Reject", "").split(",") if w.strip()]

    def __filter(self, record):
        """Check if accept or reject.

        :returns: boolean for filter value
        """
        msgPs = [record.args[0], record.varmessage]
        # if accept is empty, we accept everything and only reject will matter
        accepted = not self._accept or any(w in msg for w in self._accept for msg in msgPs)
        if not accepted:  # if accepted is False we just stop here
            return False
        rejected = any(w in msg for w in self._reject for msg in msgPs)
        return not rejected  # accepted must be True

    def filter(self, record):
        """Filter records based on the path of the logger."""
        return self.__filter(record)



#!/usr/bin/env python
"""
Script to apply update to all or some dirac servers and restart them
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from io import open

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch(
        "",
        "hosts=",
        "Comma separated list of hosts or file containing row wise list of hosts"
        " targeted for update (leave empty for all)",
    )
    Script.registerSwitch(
        "",
        "excludeHosts=",
        "Comma separated list of hosts or file containing row wise list of hosts" " excluded from update",
    )
    Script.registerSwitch("", "retry=", "Number of retry attempts on hosts that have failed to update")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("version: version of DIRAC you want to update to")

    Script.parseCommandLine(ignoreErrors=False)

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    version = Script.getPositionalArgs(group=True)
    retry = 0
    hosts = []
    excludeHosts = []

    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "hosts":
            hosts = switch[1]
        if switch[0] == "excludeHosts":
            excludeHosts = switch[1]
        if switch[0] == "retry":
            retry = int(switch[1])

    # Parse targeted hosts
    try:
        with open(hosts, "r") as f:
            hosts = f.read().splitlines()
            hosts = [str(host) for host in hosts]
    except Exception:
        pass

    if not isinstance(hosts, list):
        hosts = hosts.split(",")

    # Parse excluded hosts
    try:
        with open(excludeHosts, "r") as f:
            excludeHosts = f.read().splitlines()
            excludeHosts = [str(host) for host in excludeHosts]
    except Exception:
        pass

    if not isinstance(excludeHosts, list):
        excludeHosts = excludeHosts.split(",")

    from concurrent.futures import ThreadPoolExecutor, as_completed

    from DIRAC import S_OK, S_ERROR
    from DIRAC import gLogger
    from DIRAC.FrameworkSystem.Client.SystemAdministratorClient import SystemAdministratorClient
    from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient

    def getListOfHosts():
        """
        Return the list of all hosts that constitute a DIRAC instance
        """
        client = ComponentMonitoringClient()
        result = client.getHosts({}, False, False)
        if result["OK"]:
            hosts = [host["HostName"] for host in result["Value"]]
            return S_OK(hosts)
        return S_ERROR("Cannot get list of hosts: %s" % result["Message"])

    def parseHostname(hostName):
        """
        Separate the hostname from the port

        :param str hostName: hostname you want to parse
        """
        hostList = hostName.split(":")
        host = hostList[0]
        if len(hostList) == 2:
            port = hostList[1]
        else:
            port = None
        return host, port

    def updateHost(hostName, version):
        """
        Apply update to specific host

        :param str hostName: name of the host you want to update
        :param str version: version vArBpC you want to update to
        """
        host, port = parseHostname(hostName)

        client = SystemAdministratorClient(host, port)
        result = client.ping()
        if not result["OK"]:
            gLogger.error("Cannot connect to %s" % host)
            return result

        # If the server is already running python 3,
        # convert the version to be installed to python3
        existingVersion = result["Value"]["version"]
        if DIRAC.isPy3VersionNumber(existingVersion):
            version = DIRAC.convertToPy3VersionNumber(version)

        gLogger.notice(
            "Initiating software update of %s to %s, this can take a while, please be patient ..." % (host, version)
        )

        result = client.updateSoftware(version, "", "", timeout=600)
        if not result["OK"]:
            return result
        return S_OK()

    def updateHosts(version, hosts=[], excludeHosts=[]):
        """
        Apply update to all hosts

        :param str version: version vArBpC you want to update to
        :param list[str] hosts: list of hosts to be updated, defaults to []
        :param list[str] excludeHosts: list of hosts to be excluded from updated, defaults to []
        """
        if not hosts:
            result = getListOfHosts()
            if not result["OK"]:
                return result
            hosts = result["Value"]

        # Subtract the excluded hosts
        hosts = [host for host in hosts if host not in excludeHosts]

        updateSuccess = []
        updateFail = []

        executor = ThreadPoolExecutor(max_workers=len(hosts))
        futureUpdate = {executor.submit(updateHost, host, version): host for host in hosts}
        for future in as_completed(futureUpdate):
            host = futureUpdate[future]
            result = future.result()
            if result["OK"]:
                updateSuccess.append(host)
            else:
                updateFail.append(host)

        if not updateFail:
            gLogger.notice("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            gLogger.notice("!!! Successfully updated all hosts !!!")
            gLogger.notice("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            return S_OK([updateSuccess, updateFail])
        if not updateSuccess:
            gLogger.notice("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            gLogger.notice("XXXXX Failed to update all hosts XXXXX")
            gLogger.notice("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            return S_OK([updateSuccess, updateFail])
        gLogger.notice("X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!")
        gLogger.notice("X!X!X Partially updated hosts X!X!X!")
        gLogger.notice("Succeeded to update:")
        for host in updateSuccess:
            gLogger.notice(" + %s" % host)
        gLogger.notice("Failed to update:")
        for host in updateFail:
            gLogger.notice(" - %s" % host)
        gLogger.notice("X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!X!")
        return S_OK([updateSuccess, updateFail])

    def restartHost(hostName):
        """
        Restart all systems and components of a host

        :param str hostName: name of the host you want to restart
        """
        host, port = parseHostname(hostName)

        gLogger.notice("Pinging %s ..." % host)

        client = SystemAdministratorClient(host, port)
        result = client.ping()
        if not result["OK"]:
            gLogger.error("Could not connect to %s: %s" % (host, result["Message"]))
            return result
        gLogger.notice("Host %s is active" % host)

        gLogger.notice("Initiating restart of all systems and components")
        # This restart call will always return S_ERROR because of SystemAdministrator restart
        # Connection will be lost to the host
        result = client.restartComponent("*", "*")
        if result["Message"] == "Peer closed connection":
            gLogger.notice("Restarted all systems on %s : connection to SystemAdministrator lost" % host)
            return S_OK(result["Message"])
        gLogger.error("Received unxpected message: %s" % result["Message"])
        return result

    def updateInstance(version, hosts, excludeHosts, retry):
        """
        Update each server of an instance and restart them

        :param str version: version vArBpC you want to update to
        :param list[str] hosts: list of hosts to be updated
        :param list[str] excludeHosts: list of hosts to be excluded from update
        :param int retry: number of retry attempts on hosts that have failed to update
        """
        result = updateHosts(version, hosts, excludeHosts)
        if not result["OK"]:
            return result

        updateSuccess = result["Value"][0]
        updateFail = result["Value"][1]
        restartSuccess = []
        restartFail = []
        for host in updateSuccess:
            result = restartHost(host)
            if result["OK"]:
                restartSuccess.append(host)
            else:
                restartFail.append(host)

        if not restartFail and not updateFail:
            return S_OK("Successfully updated and restarted all hosts")

        gLogger.notice("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
        gLogger.notice("XXXXX There were problems in the update process XXXXX")
        gLogger.notice("Succeeded to update:")
        for host in updateSuccess:
            gLogger.notice(" + %s" % host)
        gLogger.notice("Succeeded to restart:")
        for host in restartSuccess:
            gLogger.notice(" + %s" % host)
        gLogger.notice("Failed to update:")
        for host in updateFail:
            gLogger.notice(" - %s" % host)
        gLogger.notice("Failed to restart:")
        for host in restartFail:
            gLogger.notice(" - %s" % host)
        gLogger.notice("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")

        if retry > 0:
            retryHosts = list(set(updateFail + restartFail))
            gLogger.notice("Retrying update on (%s attempts remaining):" % retry)
            for host in retryHosts:
                gLogger.notice(" - %s" % host)
            gLogger.notice("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            return updateInstance(version, retryHosts, excludeHosts, retry - 1)

        return S_ERROR("Update failed!")

    result = updateInstance(version, hosts, excludeHosts, retry)
    if not result["OK"]:
        gLogger.fatal(result["Message"])
        DIRAC.exit(1)
    gLogger.notice(result["Value"])
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Generate a matrix of protocols used between SEs for FTS transfers.
The output is a CSV file containing a matrix source/destination.
The value of each cell is of the form "proto1/proto2 ([protoA,protoB,...])"
proto1/proto2 are the protocols that would really be given to FTS for source and dest urls
protoA,protoB,etc are the sorted list of protocols that would be attempted for thrid party copy by DIRAC

By default, all the SEs are taken into account, but the matrix is factorized by using baseSEs.

Suppose you have the following in your CS::

  StorageElementBases{
    IN2P3-Disk
  }
  StorageElements{
    IN2P3-DST{
      BaseSE = IN2P3-Disk
    }
    IN2P3-User{
      BaseSE = IN2P3-Disk
    }
    AnotherDisk{
    }
  }

You can have the following combinations::

  DIRAC-PROD>dirac-dms-protocol-matrix
  Using sources: IN2P3-Disk, AnotherDisk
  Using target: IN2P3-Disk, AnotherDisk

  DIRAC-PROD>dirac-dms-protocol-matrix --FromSE=IN2P3-User
  Using sources: IN2P3-User
  Using target: IN2P3-Disk, AnotherDisk

  DIRAC-PROD>dirac-dms-protocol-matrix --FromSE=IN2P3-User --Bidirection
  Using sources: IN2P3-User
  Using target: IN2P3-User
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import csv
import six
from collections import defaultdict

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("", "FromSE=", "SE1[,SE2,...]")
    Script.registerSwitch("", "TargetSE=", "SE1[,SE2,...]")
    Script.registerSwitch("", "OutputFile=", "CSV output file (default /tmp/protocol-matrix.csv)")
    Script.registerSwitch("", "Bidirection", "If FromSE or TargetSE are specified, make a square matrix ")
    Script.registerSwitch("", "FTSOnly", "Only display the protocols sent to FTS")
    Script.registerSwitch("", "ExcludeSE=", "SEs to not take into account for the matrix")

    Script.parseCommandLine()
    from DIRAC import gConfig, gLogger, S_ERROR
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
    from DIRAC.DataManagementSystem.private.FTS3Utilities import getFTS3Plugin
    from DIRAC.Resources.Storage.StorageElement import StorageElement
    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup

    fromSE = []
    targetSE = []
    excludeSE = []
    outputFile = "/tmp/protocol-matrix.csv"
    bidirection = False
    ftsOnly = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "FromSE":
            fromSE = switch[1].split(",")
        elif switch[0] == "TargetSE":
            targetSE = switch[1].split(",")
        elif switch[0] == "ExcludeSE":
            excludeSE = switch[1].split(",")
        elif switch[0] == "OutputFile":
            outputFile = switch[1]
        elif switch[0] == "Bidirection":
            bidirection = True
        elif switch[0] == "FTSOnly":
            ftsOnly = True

    fts3Plugin = getFTS3Plugin()
    thirdPartyProtocols = DMSHelpers().getThirdPartyProtocols()

    # List all the BaseSE
    seBases = gConfig.getSections("/Resources/StorageElementBases")["Value"]
    # construct a dict { baseSE : <inherited storages>}
    seForSeBases = {}

    allSEs = gConfig.getSections("/Resources/StorageElements/")["Value"]

    # Remove the SEs that we want to exclude
    allSEs = set(allSEs) - set(excludeSE)

    # We go through all the SEs and fill in the seForSEBases dict.
    # Basically, at the end of the loop, the dict will contain
    # for each baseSE an entry corresponding to one real storage (the first one)
    # and itself for each real non inherited SE
    for se in allSEs:
        baseSE = gConfig.getOption("/Resources/StorageElements/%s/BaseSE" % se).get("Value")
        if baseSE:
            if baseSE not in seForSeBases:
                seForSeBases[baseSE] = se
        else:
            # If no baseSE, we put self
            seForSeBases[se] = se

    # Now let's take into account what source and destination we want.

    # If the user did not specify source or dest, generate everything
    if not fromSE and not targetSE:
        fromSE = list(seForSeBases)
        targetSE = list(seForSeBases)
    else:  # he specified at least source of dest

        # if bidirection, source and target should be the same
        if bidirection:
            if not fromSE and targetSE:  # we gave target, but no source
                fromSE = targetSE
            elif fromSE and not targetSE:  # we gave source but no target
                targetSE = fromSE
            elif fromSE and targetSE:  # we gave both
                fromSE = targetSE = list(set(fromSE + targetSE))

        else:  # no bidirection
            # he specified a targetSE
            if not fromSE:
                fromSE = list(seForSeBases)
            elif not targetSE:
                targetSE = list(seForSeBases)

    fromSE = sorted(fromSE)
    targetSE = sorted(targetSE)

    gLogger.notice("Using sources: %s" % ",".join(fromSE))
    gLogger.notice("Using target: %s" % ",".join(targetSE))

    # Now we construct the SE object for each SE that we want to appear
    ses = {}
    for se in set(fromSE + targetSE):
        ses[se] = StorageElement(seForSeBases.get(se, se))

    ret = getVOfromProxyGroup()
    if not ret["OK"] or not ret.get("Value", ""):
        gLogger.error("Aborting, Bad Proxy:", ret.get("Message", "Proxy does not belong to a VO!"))
        exit(1)
    vo = ret["Value"]
    gLogger.notice("Using the Virtual Organization:", vo)
    # dummy LFN, still has to follow lfn convention
    lfn = "/%s/toto.xml" % vo

    # Create a matrix of protocol src/dest

    tpMatrix = defaultdict(dict)

    # For each source and destination, generate the url pair, and the compatible third party protocols
    for src, dst in ((x, y) for x in fromSE for y in targetSE):
        try:
            fts3TpcProto = fts3Plugin.selectTPCProtocols(sourceSEName=src, destSEName=dst)
            res = ses[dst].generateTransferURLsBetweenSEs(lfn, ses[src], fts3TpcProto)
        except ValueError as e:
            res = S_ERROR(str(e))
        if not res["OK"]:
            surls = "Error"
            gLogger.notice("Could not generate transfer URLS", "src:%s, dst:%s, error:%s" % (src, dst, res["Message"]))
        else:
            # We only keep the protocol part of the url
            surls = "/".join(res["Value"]["Protocols"])

        # Add also the third party protocols
        proto = ",".join(ses[dst].negociateProtocolWithOtherSE(ses[src], thirdPartyProtocols)["Value"])
        if ftsOnly:
            tpMatrix[src][dst] = "%s" % surls
        else:
            tpMatrix[src][dst] = "%s (%s)" % (surls, proto)
        gLogger.verbose("%s -> %s: %s" % (src, dst, surls))
        gLogger.verbose("%s -> %s: %s" % (src, dst, proto))

    # Write the matrix in the file
    with open(outputFile, "wb" if six.PY2 else "w") as csvfile:
        csvWriter = csv.writer(csvfile, delimiter=";", quoting=csv.QUOTE_MINIMAL)

        csvWriter.writerow(["src/dst"] + targetSE)

        for src in fromSE:
            srcRow = [src]
            for dst in targetSE:
                srcRow.append(tpMatrix[src].get(dst, "NA"))
            csvWriter.writerow(srcRow)
    gLogger.notice("Wrote Matrix to", outputFile)


if __name__ == "__main__":
    main()



""" Unit tests for PathFinder only for functions that I added
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
from diraccfg import CFG

from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.ConfigurationSystem.Client.Helpers import Operations
from DIRAC.ConfigurationSystem.private.ConfigurationData import ConfigurationData

localCFGData = ConfigurationData(False)
mergedCFG = CFG()
mergedCFG.loadFromBuffer(
    """
DIRAC
{
  Setup=TestSetup
  Setups
  {
    TestSetup
    {
      WorkloadManagement=MyWM
    }
  }
}
Systems
{
  WorkloadManagement
  {
    MyWM
    {
      URLs
      {
        Service1 = dips://server1:1234/WorkloadManagement/Service1
        Service2 = dips://$MAINSERVERS$:5678/WorkloadManagement/Service2
      }
      FailoverURLs
      {
        Service2 = dips://failover1:5678/WorkloadManagement/Service2
        Service2 += dips://failover2:5678/WorkloadManagement/Service2
      }
    }
  }
}
Operations{
  Defaults
  {
    MainServers = gw1, gw2
  }
}
"""
)
localCFGData.localCFG = mergedCFG
localCFGData.remoteCFG = mergedCFG
localCFGData.mergedCFG = mergedCFG
localCFGData.generateNewVersion()


@pytest.fixture
def pathFinder(monkeypatch):
    monkeypatch.setattr(PathFinder, "gConfigurationData", localCFGData)
    monkeypatch.setattr(Operations, "gConfigurationData", localCFGData)
    return PathFinder


def test_getDIRACSetup(pathFinder):
    """Test getDIRACSetup"""
    assert pathFinder.getDIRACSetup() == "TestSetup"


@pytest.mark.parametrize(
    "system, componentName, setup, componentType, result",
    [
        (
            "WorkloadManagement/SandboxStoreHandler",
            False,
            False,
            "Services",
            "/Systems/WorkloadManagement/MyWM/Services/SandboxStoreHandler",
        ),
        (
            "WorkloadManagement",
            "SandboxStoreHandler",
            False,
            "Services",
            "/Systems/WorkloadManagement/MyWM/Services/SandboxStoreHandler",
        ),
        # tricky case one could expect that if entity string is wrong
        # than some kind of error will be returned, but it is not the case
        (
            "WorkloadManagement/SimpleLogConsumer",
            False,
            False,
            "NonRonsumersNon",
            "/Systems/WorkloadManagement/MyWM/NonRonsumersNon/SimpleLogConsumer",
        ),
    ],
)
def test_getComponentSection(pathFinder, system, componentName, setup, componentType, result):
    """Test getComponentSection"""
    assert pathFinder.getComponentSection(system, componentName, setup, componentType) == result


@pytest.mark.parametrize(
    "system, setup, result",
    [
        ("WorkloadManagement", False, "/Systems/WorkloadManagement/MyWM/URLs"),
        ("WorkloadManagement", "TestSetup", "/Systems/WorkloadManagement/MyWM/URLs"),
    ],
)
def test_getSystemURLSection(pathFinder, system, setup, result):
    assert pathFinder.getSystemURLs(system, setup)


@pytest.mark.parametrize(
    "serviceName, service, result",
    [
        (
            "WorkloadManagement/Service1",
            None,
            {"dips://server1:1234/WorkloadManagement/Service1"},
        ),
        (
            "WorkloadManagement",
            "Service1",
            {"dips://server1:1234/WorkloadManagement/Service1"},
        ),
        (
            "WorkloadManagement",
            "Service2",
            {
                "dips://gw1:5678/WorkloadManagement/Service2",
                "dips://gw2:5678/WorkloadManagement/Service2",
            },
        ),
    ],
)
def test_getServiceURL(pathFinder, serviceName, service, result):
    assert set(List.fromChar(pathFinder.getServiceURL(serviceName, service=service))) == result


@pytest.mark.parametrize(
    "serviceName, service, result",
    [
        ("WorkloadManagement/Service1", None, ""),
        ("WorkloadManagement", "Service1", ""),
        (
            "WorkloadManagement",
            "Service2",
            "dips://failover1:5678/WorkloadManagement/Service2,dips://failover2:5678/WorkloadManagement/Service2",
        ),
    ],
)
def test_getServiceFailoverURL(pathFinder, serviceName, service, result):
    """Test getServiceFailoverURL"""
    assert pathFinder.getServiceFailoverURL(serviceName, service=service) == result


@pytest.mark.parametrize(
    "serviceName, service, failover, result",
    [
        (
            "WorkloadManagement/Service1",
            None,
            False,
            {"dips://server1:1234/WorkloadManagement/Service1"},
        ),
        (
            "WorkloadManagement",
            "Service1",
            False,
            {"dips://server1:1234/WorkloadManagement/Service1"},
        ),
        (
            "WorkloadManagement",
            "Service2",
            False,
            {
                "dips://gw1:5678/WorkloadManagement/Service2",
                "dips://gw2:5678/WorkloadManagement/Service2",
            },
        ),
        (
            "WorkloadManagement",
            "Service1",
            True,
            {"dips://server1:1234/WorkloadManagement/Service1"},
        ),
        (
            "WorkloadManagement",
            "Service2",
            True,
            {
                "dips://gw1:5678/WorkloadManagement/Service2",
                "dips://gw2:5678/WorkloadManagement/Service2",
                "dips://failover1:5678/WorkloadManagement/Service2",
                "dips://failover2:5678/WorkloadManagement/Service2",
            },
        ),
    ],
)
def test_getServiceURLs(pathFinder, serviceName, service, failover, result):
    """Test getServiceURLs"""
    assert set(pathFinder.getServiceURLs(serviceName, service=service, failover=failover)) == result


@pytest.mark.parametrize(
    "system, setup, failover, result",
    [
        (
            "WorkloadManagement",
            None,
            False,
            {
                "Service1": {"dips://server1:1234/WorkloadManagement/Service1"},
                "Service2": {
                    "dips://gw1:5678/WorkloadManagement/Service2",
                    "dips://gw2:5678/WorkloadManagement/Service2",
                },
            },
        ),
        (
            "WorkloadManagement",
            None,
            True,
            {
                "Service1": {"dips://server1:1234/WorkloadManagement/Service1"},
                "Service2": {
                    "dips://gw1:5678/WorkloadManagement/Service2",
                    "dips://gw2:5678/WorkloadManagement/Service2",
                    "dips://failover1:5678/WorkloadManagement/Service2",
                    "dips://failover2:5678/WorkloadManagement/Service2",
                },
            },
        ),
    ],
)
def test_getSystemURLs(pathFinder, system, setup, failover, result):
    sysDict = pathFinder.getSystemURLs(system, setup=setup, failover=failover)
    for service in sysDict:
        assert set(sysDict[service]) == result[service]


@pytest.mark.parametrize(
    "serviceURL, system, service, result",
    [
        (
            "dips://server.com:1234/WorkloadManagement/Service1",
            None,
            None,
            "dips://server.com:1234/WorkloadManagement/Service1",
        ),
        (
            "dips://server.com:1234/",
            "WorkloadManagement",
            "Service1",
            "dips://server.com:1234/WorkloadManagement/Service1",
        ),
        (
            "dips://server.com:1234",
            "WorkloadManagement",
            "Service1",
            "dips://server.com:1234/WorkloadManagement/Service1",
        ),
        ("dips://server.com:1234/", "WorkloadManagement", None, "raise:path"),
        ("dips://server.com/WorkloadManagement/Service1", None, None, "raise:port"),
        (
            "https://server.com/WorkloadManagement/Service1",
            None,
            None,
            "https://server.com:443/WorkloadManagement/Service1",
        ),
        (
            "http://server.com/WorkloadManagement/Service1",
            None,
            None,
            "http://server.com:80/WorkloadManagement/Service1",
        ),
    ],
)
def test_checkComponentURL(pathFinder, serviceURL, system, service, result):
    try:
        pathFinderResult = pathFinder.checkComponentURL(serviceURL, system, service, pathMandatory=True)
        assert pathFinderResult == result
    except RuntimeError as e:
        assert result.split(":")[1] in repr(e)



########################################################################
# File: ReqProxyHandler.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/06/04 13:18:41
########################################################################

"""
:mod: RequestProxyHandler

.. module: ReqtProxyHandler
  :synopsis: ReqProxy service

.. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

Careful with that axe, Eugene! Some 'transfer' requests are using local fs
and they never should be forwarded to the central RequestManager.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ReqProxy
  :end-before: ##END
  :dedent: 2
  :caption: ReqProxy options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# #
# @file RequestProxyHandler.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/07/20 13:18:58
# @brief Definition of RequestProxyHandler class.

# # imports
import os

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

import json
import six

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.RequestManagementSystem.Client.Request import Request


def initializeReqProxyHandler(serviceInfo):
    """init RequestProxy handler

    :param serviceInfo: whatever
    """
    gLogger.info("Initalizing ReqProxyHandler")
    gThreadScheduler.addPeriodicTask(120, ReqProxyHandler.sweeper)
    return S_OK()


########################################################################


class ReqProxyHandler(RequestHandler):
    """
    .. class:: ReqProxyHandler

    :param RPCCLient requestManager: a RPCClient to RequestManager
    :param str cacheDir: os.path.join( workDir, "requestCache" )
    """

    __requestManager = None
    __cacheDir = None

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialize handler"""
        gLogger.notice("CacheDirectory: %s" % cls.cacheDir())
        gMonitor.registerActivity("reqSwept", "Request successfully swept", "ReqProxy", "Requests/min", gMonitor.OP_SUM)
        gMonitor.registerActivity("reqFailed", "Request forward failed", "ReqProxy", "Requests/min", gMonitor.OP_SUM)
        gMonitor.registerActivity("reqReceived", "Request received", "ReqProxy", "Requests/min", gMonitor.OP_SUM)
        cls.sweepSize = getServiceOption(serviceInfoDict, "SweepSize", 10)
        gLogger.notice("SweepSize: %s" % cls.sweepSize)
        return S_OK()

    @classmethod
    def requestManager(cls):
        """get request manager"""
        if not cls.__requestManager:
            cls.__requestManager = RPCClient("RequestManagement/ReqManager")
        return cls.__requestManager

    @classmethod
    def cacheDir(cls):
        """get cache dir"""
        if not cls.__cacheDir:
            cls.__cacheDir = os.path.abspath("requestCache")
            if not os.path.exists(cls.__cacheDir):
                os.mkdir(cls.__cacheDir)
        return cls.__cacheDir

    @classmethod
    def sweeper(cls):
        """move cached request to the central request manager

        :param self: self reference
        """
        cacheDir = cls.cacheDir()

        # # cache dir empty?
        if not os.listdir(cacheDir):
            gLogger.always("sweeper: CacheDir %s is empty, nothing to do" % cacheDir)
            return S_OK()
        else:
            # # read <sweepSize> cache dir files, the oldest first
            cachedRequests = [
                os.path.abspath(requestFile)
                for requestFile in sorted(
                    filter(
                        os.path.isfile, [os.path.join(cacheDir, requestName) for requestName in os.listdir(cacheDir)]
                    ),
                    key=os.path.getctime,
                )
            ][: cls.sweepSize]
            # # set cached requests to the central RequestManager
            for cachedFile in cachedRequests:
                # # break if something went wrong last time
                try:
                    requestJSON = "".join(open(cachedFile, "r").readlines())
                    cachedRequest = json.loads(requestJSON)
                    cachedName = cachedRequest.get("RequestName", "***UNKNOWN***")
                    putRequest = cls.requestManager().putRequest(requestJSON)
                    if not putRequest["OK"]:
                        gLogger.error(
                            "sweeper: unable to set request %s @ ReqManager: %s" % (cachedName, putRequest["Message"])
                        )
                        gMonitor.addMark("reqFailed", 1)

                        continue
                    gLogger.info("sweeper: successfully put request '%s' @ ReqManager" % cachedName)
                    gMonitor.addMark("reqSwept", 1)
                    os.unlink(cachedFile)
                except Exception as error:
                    gMonitor.addMark("reqFailed", 1)
                    gLogger.exception("sweeper: hit by exception", lException=error)

            return S_OK()

    def __saveRequest(self, requestName, requestJSON):
        """save request string to the working dir cache

        :param self: self reference
        :param str requestName: request name
        :param str requestJSON:  request serialized to JSON format
        """
        try:
            requestFile = os.path.join(self.cacheDir(), md5(requestJSON.encode()).hexdigest())
            with open(requestFile, "w+") as request:
                request.write(requestJSON)
            return S_OK(requestFile)
        except OSError as error:
            err = "unable to dump %s to cache file: %s" % (requestName, str(error))
            gLogger.exception(err)
            return S_ERROR(err)

    types_getStatus = []

    def export_getStatus(self):
        """get number of requests in cache"""
        try:
            cachedRequests = len(os.listdir(self.cacheDir()))
        except OSError as error:
            err = "getStatus: unable to list cache dir contents: %s" % str(error)
            gLogger.exception(err)
            return S_ERROR(err)
        return S_OK(cachedRequests)

    types_putRequest = [six.string_types]

    def export_putRequest(self, requestJSON):
        """forward request from local RequestDB to central RequestManager

        :param self: self reference
        :param str requestType: request type
        """

        gMonitor.addMark("reqReceived", 1)

        requestDict = json.loads(requestJSON)
        requestName = requestDict.get("RequestID", requestDict.get("RequestName", "***UNKNOWN***"))
        gLogger.info("putRequest: got request '%s'" % requestName)

        # We only need the object to check the authorization
        request = Request(requestDict)
        # Check whether the credentials in the Requests are correct and allowed to be set
        isAuthorized = RequestValidator.setAndCheckRequestOwner(request, self.getRemoteCredentials())

        if not isAuthorized:
            return S_ERROR(DErrno.ENOAUTH, "Credentials in the requests are not allowed")

        forwardable = self.__forwardable(requestDict)
        if not forwardable["OK"]:
            gLogger.warn("putRequest: %s" % forwardable["Message"])

        setRequest = self.requestManager().putRequest(requestJSON)
        if not setRequest["OK"]:
            gLogger.error(
                "setReqeuest: unable to set request '%s' @ RequestManager: %s" % (requestName, setRequest["Message"])
            )
            # # put request to the request file cache
            save = self.__saveRequest(requestName, requestJSON)
            if not save["OK"]:
                gLogger.error("setRequest: unable to save request to the cache: %s" % save["Message"])
                return save
            gLogger.info("setRequest: %s is saved to %s file" % (requestName, save["Value"]))
            return S_OK({"set": False, "saved": True})

        gLogger.info("setRequest: request '%s' has been set to the ReqManager" % (requestName))
        return S_OK({"set": True, "saved": False})

    @staticmethod
    def __forwardable(requestDict):
        """check if request if forwardable

        The sub-request of type transfer:putAndRegister, removal:physicalRemoval and removal:reTransfer are
        definitely not, they should be executed locally, as they are using local fs.

        :param str requestJSON: serialized request
        """
        operations = requestDict.get("Operations", [])
        for operationDict in operations:
            if operationDict.get("Type", "") in ("PutAndRegister", "PhysicalRemoval", "ReTransfer"):
                return S_ERROR(
                    DErrno.ERMSUKN, "found operation '%s' that cannot be forwarded" % operationDict.get("Type", "")
                )
        return S_OK()

    types_listCacheDir = []

    def export_listCacheDir(self):
        """List the content of the Cache directory

        :returns: list of file
        """
        cacheDir = self.cacheDir()
        try:
            dirContent = os.listdir(cacheDir)
            return S_OK(dirContent)
        except OSError as e:
            return S_ERROR(DErrno.ERMSUKN, "Error listing %s: %s" % (cacheDir, repr(e)))

    types_showCachedRequest = [six.string_types]

    def export_showCachedRequest(self, filename):
        """Show the request cached in the given file"""
        fullPath = None
        try:
            fullPath = os.path.join(self.cacheDir(), filename)
            with open(fullPath, "r") as cacheFile:
                requestJSON = "".join(cacheFile.readlines())
                return S_OK(requestJSON)
        except Exception as e:
            return S_ERROR(DErrno.ERMSUKN, "Error showing cached request %s: %s" % (fullPath, repr(e)))



""" Test class for SiteDirector
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

# imports
import datetime
import pytest
from mock import MagicMock

from DIRAC import gLogger

# sut
from DIRAC.WorkloadManagementSystem.Agent.SiteDirector import SiteDirector

mockAM = MagicMock()
mockGCReply = MagicMock()
mockGCReply.return_value = "TestSetup"
mockOPSObject = MagicMock()
mockOPSObject.getValue.return_value = "123"
mockOPSReply = MagicMock()
mockOPSReply.return_value = "123"

mockOPS = MagicMock()
mockOPS.return_value = mockOPSObject
# mockOPS.Operations = mockOPSObject
mockPM = MagicMock()
mockPM.requestToken.return_value = {"OK": True, "Value": ("token", 1)}
mockPMReply = MagicMock()
mockPMReply.return_value = {"OK": True, "Value": ("token", 1)}

mockCSGlobalReply = MagicMock()
mockCSGlobalReply.return_value = "TestSetup"
mockResourcesReply = MagicMock()
mockResourcesReply.return_value = {"OK": True, "Value": ["x86_64-slc6", "x86_64-slc5"]}

mockPilotAgentsDB = MagicMock()
mockPilotAgentsDB.setPilotStatus.return_value = {"OK": True}

gLogger.setLevel("DEBUG")


@pytest.fixture
def sd(mocker):
    """mocker for SiteDirector"""
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.SiteDirector.AgentModule.__init__")
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.SiteDirector.gConfig.getValue", side_effect=mockGCReply)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.SiteDirector.Operations", side_effect=mockOPS)
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.SiteDirector.gProxyManager.requestToken", side_effect=mockPMReply
    )
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.SiteDirector.AgentModule", side_effect=mockAM)
    sd = SiteDirector()
    sd.log = gLogger
    sd.am_getOption = mockAM
    sd.log.setLevel("DEBUG")
    sd.rpcMatcher = MagicMock()
    sd.rssClient = MagicMock()
    sd.workingDirectory = ""
    sd.queueDict = {
        "aQueue": {
            "Site": "LCG.CERN.cern",
            "CEName": "aCE",
            "CEType": "SSH",
            "QueueName": "aQueue",
            "ParametersDict": {
                "CPUTime": 12345,
                "Community": "lhcb",
                "OwnerGroup": ["lhcb_user"],
                "Setup": "LHCb-Production",
                "Site": "LCG.CERN.cern",
            },
        }
    }
    return sd


def test__getPilotOptions(sd):
    """Testing SiteDirector()._getPilotOptions()"""
    res = sd._getPilotOptions("aQueue")
    assert set(["-S TestSetup", "-V 123", "-l 123", "-n LCG.CERN.cern"]) <= set(res)


@pytest.mark.parametrize(
    "mockMatcherReturnValue, expected, anyExpected, sitesExpected",
    [
        ({"OK": False, "Message": "boh"}, False, True, set()),
        ({"OK": True, "Value": None}, False, True, set()),
        ({"OK": True, "Value": {"1": {"Jobs": 10}, "2": {"Jobs": 20}}}, True, True, set()),
        ({"OK": True, "Value": {"1": {"Jobs": 10, "Sites": ["Site1"]}, "2": {"Jobs": 20}}}, True, True, set(["Site1"])),
        (
            {"OK": True, "Value": {"1": {"Jobs": 10, "Sites": ["Site1", "Site2"]}, "2": {"Jobs": 20}}},
            True,
            True,
            set(["Site1", "Site2"]),
        ),
        (
            {
                "OK": True,
                "Value": {"1": {"Jobs": 10, "Sites": ["Site1", "Site2"]}, "2": {"Jobs": 20, "Sites": ["Site1"]}},
            },
            True,
            False,
            set(["Site1", "Site2"]),
        ),
        (
            {
                "OK": True,
                "Value": {"1": {"Jobs": 10, "Sites": ["Site1", "Site2"]}, "2": {"Jobs": 20, "Sites": ["ANY"]}},
            },
            True,
            False,
            {"Site1", "Site2", "ANY"},
        ),
        (
            {
                "OK": True,
                "Value": {"1": {"Jobs": 10, "Sites": ["Site1", "Site2"]}, "2": {"Jobs": 20, "Sites": ["ANY", "Site3"]}},
            },
            True,
            False,
            {"Site1", "Site2", "Site3", "ANY"},
        ),
        (
            {
                "OK": True,
                "Value": {"1": {"Jobs": 10, "Sites": ["Site1", "Site2"]}, "2": {"Jobs": 20, "Sites": ["Any", "Site3"]}},
            },
            True,
            False,
            {"Site1", "Site2", "Site3", "Any"},
        ),
        (
            {
                "OK": True,
                "Value": {
                    "1": {"Jobs": 10, "Sites": ["Site1", "Site2"]},
                    "2": {"Jobs": 20, "Sites": ["NotAny", "Site2"]},
                },
            },
            True,
            False,
            {"Site1", "Site2", "NotAny"},
        ),
    ],
)
def test__ifAndWhereToSubmit(sd, mockMatcherReturnValue, expected, anyExpected, sitesExpected):
    """Testing SiteDirector()._ifAndWhereToSubmit()"""
    sd.matcherClient = MagicMock()
    sd.matcherClient.getMatchingTaskQueues.return_value = mockMatcherReturnValue
    res = sd._ifAndWhereToSubmit()
    assert res[0] == expected
    if res[0]:
        assert res == (expected, anyExpected, sitesExpected, set())


def test__allowedToSubmit(sd):
    """Testing SiteDirector()._allowedToSubmit()"""
    submit = sd._allowedToSubmit("aQueue", True, set(["LCG.CERN.cern"]), set())
    assert submit is False

    sd.siteMaskList = ["LCG.CERN.cern", "DIRAC.CNAF.it"]
    submit = sd._allowedToSubmit("aQueue", True, set(["LCG.CERN.cern"]), set())
    assert submit is True

    sd.rssFlag = True
    submit = sd._allowedToSubmit("aQueue", True, set(["LCG.CERN.cern"]), set())
    assert submit is False

    sd.ceMaskList = ["aCE", "anotherCE"]
    submit = sd._allowedToSubmit("aQueue", True, set(["LCG.CERN.cern"]), set())
    assert submit is True


def test__submitPilotsToQueue(sd):
    """Testing SiteDirector()._submitPilotsToQueue()"""
    # Create a MagicMock that does not have the workingDirectory
    # attribute (https://cpython-test-docs.readthedocs.io/en/latest/library/unittest.mock.html#deleting-attributes)
    # This is to use the SiteDirector's working directory, not the CE one
    ceMock = MagicMock()
    del ceMock.workingDirectory

    sd.queueCECache = {"aQueue": {"CE": ceMock}}
    sd.queueSlots = {"aQueue": {"AvailableSlots": 10}}
    assert sd._submitPilotsToQueue(1, MagicMock(), "aQueue")["OK"]


@pytest.mark.parametrize(
    "pilotRefs, pilotDict, pilotCEDict, expected",
    [
        ([], {}, {}, (0, [])),
        (
            ["aPilotRef"],
            {"aPilotRef": {"Status": "Running", "LastUpdateTime": datetime.datetime(2000, 1, 1).utcnow()}},
            {},
            (0, []),
        ),
        (
            ["aPilotRef"],
            {"aPilotRef": {"Status": "Running", "LastUpdateTime": datetime.datetime(2000, 1, 1).utcnow()}},
            {"aPilotRef": "Running"},
            (0, []),
        ),
        (
            ["aPilotRef"],
            {"aPilotRef": {"Status": "Running", "LastUpdateTime": datetime.datetime(2000, 1, 1).utcnow()}},
            {"aPilotRef": "Unknown"},
            (0, []),
        ),
    ],
)
def test__updatePilotStatus(sd, pilotRefs, pilotDict, pilotCEDict, expected):
    """Testing SiteDirector()._updatePilotStatus()"""
    res = sd._updatePilotStatus(pilotRefs, pilotDict, pilotCEDict)
    assert res == expected



""" LogStatusAction

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.PolicySystem.Actions.BaseAction import BaseAction


class LogStatusAction(BaseAction):
    """
    Action that registers on the database a new entry on the <element>Status table.
    It adds or modifies if the record exists on the table.
    """

    def __init__(self, name, decisionParams, enforcementResult, singlePolicyResults, clients=None):

        super(LogStatusAction, self).__init__(name, decisionParams, enforcementResult, singlePolicyResults, clients)

        if clients is not None and "ResourceStatusClient" in clients:
            self.rsClient = clients["ResourceStatusClient"]
        else:
            self.rsClient = ResourceStatusClient()

    def run(self):
        """
        Checks it has the parameters it needs and tries to addOrModify in the
        database.
        """
        # Minor security checks

        element = self.decisionParams["element"]
        if element is None:
            return S_ERROR("element should not be None")

        name = self.decisionParams["name"]
        if name is None:
            return S_ERROR("name should not be None")

        statusType = self.decisionParams["statusType"]
        if statusType is None:
            return S_ERROR("statusType should not be None")

        status = self.enforcementResult["Status"]
        if status is None:
            return S_ERROR("status should not be None")

        elementType = self.decisionParams["elementType"]
        if elementType is None:
            return S_ERROR("elementType should not be None")

        reason = self.enforcementResult["Reason"]
        if reason is None:
            return S_ERROR("reason should not be None")

        vo = self.enforcementResult.get("VO")
        # Truncate reason to fit in database column
        reason = (reason[:508] + "..") if len(reason) > 508 else reason
        # VO = 'all' (non-VO aware policy) for a combined policy affects all VOs for a given site or resource,
        if vo == "all":
            resSelect = self.rsClient.selectStatusElement(
                element,
                "Status",
                name=name,
                statusType=None,
                vO=None,
                status=None,
                elementType=None,
                reason=None,
                dateEffective=None,
                lastCheckTime=None,
                tokenOwner="rs_svc",
                tokenExpiration=None,
                meta=None,
            )
            if not resSelect["OK"]:
                self.log.error("Could not obtain all VO rows for element: %s" % element)
                return resSelect
            voColumnIndex = resSelect["Columns"].index("VO")
            for row in resSelect["Value"]:
                vo = row[voColumnIndex]
                resLogUpdate = self.rsClient.addOrModifyStatusElement(
                    element,
                    "Status",
                    name=name,
                    statusType=statusType,
                    vO=vo,
                    status=status,
                    elementType=elementType,
                    reason=reason,
                )
                self.log.debug("Update result", resLogUpdate)

        else:
            resLogUpdate = self.rsClient.addOrModifyStatusElement(
                element,
                "Status",
                name=name,
                statusType=statusType,
                vO=vo,
                status=status,
                elementType=elementType,
                reason=reason,
            )

        return resLogUpdate


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" The StalledJobAgent hunts for stalled jobs in the Job database. Jobs in "running"
    state not receiving a heart beat signal for more than stalledTime
    seconds will be assigned the "Stalled" state.


.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN StalledJobAgent
  :end-before: ##END
  :dedent: 2
  :caption: StalledJobAgent options

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import concurrent.futures

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.AccountingSystem.Client.Types.Job import Job
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.Time import fromString, toEpoch, dateTime, second
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
from DIRAC.ConfigurationSystem.Client.PathFinder import getSystemInstance
from DIRAC.WorkloadManagementSystem.Client.WMSClient import WMSClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.DB.JobLoggingDB import JobLoggingDB
from DIRAC.WorkloadManagementSystem.Client import JobStatus, JobMinorStatus


class StalledJobAgent(AgentModule):
    """Agent for setting Running jobs Stalled, and Stalled jobs Failed. And a few more."""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        super(StalledJobAgent, self).__init__(*args, **kwargs)

        self.jobDB = None
        self.logDB = None
        self.matchedTime = 7200
        self.rescheduledTime = 600
        self.submittingTime = 300
        self.stalledJobsTolerantSites = []
        self.stalledJobsToRescheduleSites = []

    #############################################################################
    def initialize(self):
        """Sets default parameters"""
        self.jobDB = JobDB()
        self.logDB = JobLoggingDB()

        # getting parameters

        if not self.am_getOption("Enable", True):
            self.log.info("Stalled Job Agent running in disabled mode")

        wms_instance = getSystemInstance("WorkloadManagement")
        if not wms_instance:
            return S_ERROR("Can not get the WorkloadManagement system instance")
        self.stalledJobsTolerantSites = self.am_getOption("StalledJobsTolerantSites", [])
        self.stalledJobsToleranceTime = self.am_getOption("StalledJobsToleranceTime", 0)

        self.stalledJobsToRescheduleSites = self.am_getOption("StalledJobsToRescheduleSites", [])

        self.submittingTime = self.am_getOption("SubmittingTime", self.submittingTime)
        self.matchedTime = self.am_getOption("MatchedTime", self.matchedTime)
        self.rescheduledTime = self.am_getOption("RescheduledTime", self.rescheduledTime)

        wrapperSection = cfgPath("Systems", "WorkloadManagement", wms_instance, "JobWrapper")

        failedTime = self.am_getOption("FailedTimeHours", 6)
        watchdogCycle = gConfig.getValue(cfgPath(wrapperSection, "CheckingTime"), 30 * 60)
        watchdogCycle = max(watchdogCycle, gConfig.getValue(cfgPath(wrapperSection, "MinCheckingTime"), 20 * 60))
        stalledTime = self.am_getOption("StalledTimeHours", 2)
        self.log.verbose("", "StalledTime = %s cycles" % (stalledTime))
        self.stalledTime = int(watchdogCycle * (stalledTime + 0.5))
        self.log.verbose("", "FailedTime = %s cycles" % (failedTime))

        # Add half cycle to avoid race conditions
        self.failedTime = int(watchdogCycle * (failedTime + 0.5))

        self.minorStalledStatuses = (
            JobMinorStatus.STALLED_PILOT_NOT_RUNNING,
            "Stalling for more than %d sec" % self.failedTime,
        )

        # setting up the threading
        maxNumberOfThreads = self.am_getOption("MaxNumberOfThreads", 15)
        self.log.verbose("Multithreaded with %d threads" % maxNumberOfThreads)
        self.threadPoolExecutor = concurrent.futures.ThreadPoolExecutor(max_workers=maxNumberOfThreads)

        return S_OK()

    #############################################################################
    def execute(self):
        """The main agent execution method"""
        # Now we are getting what's going to be checked
        futures = []

        # 1) Queueing the jobs that might be marked Stalled
        # This is the minimum time we wait for declaring a job Stalled, therefore it is safe
        checkTime = dateTime() - self.stalledTime * second
        checkedStatuses = [JobStatus.RUNNING, JobStatus.COMPLETING]
        # Only get jobs whose HeartBeat is older than the stalledTime
        result = self.jobDB.selectJobs({"Status": checkedStatuses}, older=checkTime, timeStamp="HeartBeatTime")
        if not result["OK"]:
            self.log.error("Issue selecting %s jobs" % " & ".join(checkedStatuses), result["Message"])
        if result["Value"]:
            jobs = sorted(result["Value"])
            self.log.info(
                "%s jobs will be checked for being stalled" % " & ".join(checkedStatuses),
                "(n=%d, heartbeat before %s)" % (len(jobs), str(checkTime)),
            )
            for job in jobs:
                future = self.threadPoolExecutor.submit(self._execute, "%s:_markStalledJobs" % job)
                futures.append(future)

        # 2) fail Stalled Jobs
        result = self.jobDB.selectJobs({"Status": JobStatus.STALLED})
        if not result["OK"]:
            self.log.error("Issue selecting Stalled jobs", result["Message"])
        if result["Value"]:
            jobs = sorted(result["Value"])
            self.log.info("Jobs Stalled will be checked for failure", "(n=%d)" % len(jobs))
            for job in jobs:
                future = self.threadPoolExecutor.submit(self._execute, "%s:_failStalledJobs" % job)
                futures.append(future)

        # 3) Send accounting
        for minor in self.minorStalledStatuses:
            result = self.jobDB.selectJobs({"Status": JobStatus.FAILED, "MinorStatus": minor, "AccountedFlag": "False"})
            if not result["OK"]:
                self.log.error("Issue selecting jobs for accounting", result["Message"])
            if result["Value"]:
                jobs = result["Value"]
                self.log.info("Stalled jobs will be Accounted", "(n=%d)" % (len(jobs)))
                for job in jobs:
                    future = self.threadPoolExecutor.submit(self._execute, "%s:_sendAccounting" % job)
                    futures.append(future)

        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as exc:
                self.log.error("_execute generated an exception: %s" % exc)

        # From here on we don't use the threads

        # 4) Fail submitting jobs
        result = self._failSubmittingJobs()
        if not result["OK"]:
            self.log.error("Failed to process jobs being submitted", result["Message"])

        # 5) Kick stuck jobs
        result = self._kickStuckJobs()
        if not result["OK"]:
            self.log.error("Failed to kick stuck jobs", result["Message"])

        return S_OK()

    def finalize(self):
        """graceful finalization"""

        self.log.info("Wait for threads to get empty before terminating the agent")
        self.threadPoolExecutor.shutdown()
        self.log.info("Threads are empty, terminating the agent...")
        return S_OK()

    def _execute(self, job_Op):
        """
        Doing the actual job. This is run inside the threads
        """
        jobID, jobOp = job_Op.split(":")
        jobID = int(jobID)
        res = getattr(self, "%s" % jobOp)(jobID)
        if not res["OK"]:
            self.log.error("Failure executing %s" % jobOp, "on %d: %s" % (jobID, res["Message"]))

    #############################################################################
    def _markStalledJobs(self, jobID):
        """
        Identifies if JobID is stalled:
        running or completing without update longer than stalledTime.

        Run inside thread.
        """
        delayTime = self.stalledTime
        # Add a tolerance time for some sites if required
        if self.stalledJobsTolerantSites:
            result = self.jobDB.getJobAttribute(jobID, "site")
            if not result["OK"]:
                return result
            site = result["Value"]
            if site in self.stalledJobsTolerantSites:
                delayTime += self.stalledJobsToleranceTime
        # Check if the job is really stalled
        result = self._checkJobStalled(jobID, delayTime)
        if not result["OK"]:
            return result
        self.log.verbose("Updating status to Stalled", "for job %s" % (jobID))
        return self._updateJobStatus(jobID, JobStatus.STALLED)

    #############################################################################
    def _failStalledJobs(self, jobID):
        """
        Changes the Stalled status to Failed for jobs long in the Stalled status.

        Run inside thread.
        """

        setFailed = False
        # Check if the job pilot is lost
        result = self._getJobPilotStatus(jobID)
        if not result["OK"]:
            self.log.error("Failed to get pilot status", "for job %d: %s" % (jobID, result["Message"]))
            return result
        pilotStatus = result["Value"]
        if pilotStatus != "Running":
            setFailed = self.minorStalledStatuses[0]
        else:
            # Verify that there was no sign of life for long enough
            result = self._getLatestUpdateTime(jobID)
            if not result["OK"]:
                self.log.error("Failed to get job update time", "for job %d: %s" % (jobID, result["Message"]))
                return result
            elapsedTime = toEpoch() - result["Value"]
            if elapsedTime > self.failedTime:
                setFailed = self.minorStalledStatuses[1]

        # Set the jobs Failed, send them a kill signal in case they are not really dead
        # and send accounting info
        if setFailed:
            self._sendKillCommand(jobID)  # always returns None

            # For some sites we might want to reschedule rather than fail the jobs
            if self.stalledJobsToRescheduleSites:
                result = self.jobDB.getJobAttribute(jobID, "site")
                if not result["OK"]:
                    return result
                site = result["Value"]
                if site in self.stalledJobsToRescheduleSites:
                    return self._updateJobStatus(jobID, JobStatus.RESCHEDULED, minorStatus=setFailed)

            return self._updateJobStatus(jobID, JobStatus.FAILED, minorStatus=setFailed)

        return S_OK()

    def _getJobPilotStatus(self, jobID):
        """Get the job pilot status"""
        result = JobMonitoringClient().getJobParameter(jobID, "Pilot_Reference")
        if not result["OK"]:
            return result
        pilotReference = result["Value"].get("Pilot_Reference", "Unknown")
        if pilotReference == "Unknown":
            # There is no pilot reference, hence its status is unknown
            return S_OK("NoPilot")

        result = PilotManagerClient().getPilotInfo(pilotReference)
        if not result["OK"]:
            if DErrno.cmpError(result, DErrno.EWMSNOPILOT):
                self.log.warn("No pilot found", "for job %d: %s" % (jobID, result["Message"]))
                return S_OK("NoPilot")
            self.log.error("Failed to get pilot information", "for job %d: %s" % (jobID, result["Message"]))
            return result
        pilotStatus = result["Value"][pilotReference]["Status"]

        return S_OK(pilotStatus)

    #############################################################################
    def _checkJobStalled(self, job, stalledTime):
        """Compares the most recent of LastUpdateTime and HeartBeatTime against
        the stalledTime limit.
        """
        result = self._getLatestUpdateTime(job)
        if not result["OK"]:
            return result

        elapsedTime = toEpoch() - result["Value"]
        self.log.debug("(CurrentTime-LastUpdate) = %s secs" % (elapsedTime))
        if elapsedTime > stalledTime:
            self.log.info(
                "Job is identified as stalled", ": jobID %d with last update > %s secs ago" % (job, elapsedTime)
            )
            return S_OK()

        return S_ERROR("Job %s is running and will be ignored" % job)

    #############################################################################
    def _getLatestUpdateTime(self, job):
        """Returns the most recent of HeartBeatTime and LastUpdateTime"""
        result = self.jobDB.getJobAttributes(job, ["HeartBeatTime", "LastUpdateTime"])
        if not result["OK"] or not result["Value"]:
            self.log.error(
                "Failed to get job attributes",
                "for job %d: %s" % (job, result["Message"] if "Message" in result else "empty"),
            )
            return S_ERROR("Could not get attributes for job")

        latestUpdate = 0
        if not result["Value"]["HeartBeatTime"] or result["Value"]["HeartBeatTime"] == "None":
            self.log.verbose("HeartBeatTime is null", "for job %s" % job)
        else:
            latestUpdate = toEpoch(fromString(result["Value"]["HeartBeatTime"]))

        if not result["Value"]["LastUpdateTime"] or result["Value"]["LastUpdateTime"] == "None":
            self.log.verbose("LastUpdateTime is null", "for job %s" % job)
        else:
            latestUpdate = max(latestUpdate, toEpoch(fromString(result["Value"]["LastUpdateTime"])))

        if not latestUpdate:
            return S_ERROR("LastUpdate and HeartBeat times are null for job %s" % job)
        else:
            self.log.verbose("", "Latest update time from epoch for job %s is %s" % (job, latestUpdate))
            return S_OK(latestUpdate)

    #############################################################################
    def _updateJobStatus(self, job, status, minorStatus=None):
        """This method updates the job status in the JobDB"""

        if not self.am_getOption("Enable", True):
            return S_OK("Disabled")

        toRet = S_OK()

        self.log.debug("self.jobDB.setJobAttribute(%s,'Status','%s',update=True)" % (job, status))
        result = self.jobDB.setJobAttribute(job, "Status", status, update=True)
        if not result["OK"]:
            self.log.error("Failed setting Status", "%s for job %d: %s" % (status, job, result["Message"]))
            toRet = result
        if minorStatus:
            self.log.debug("self.jobDB.setJobAttribute(%s,'MinorStatus','%s',update=True)" % (job, minorStatus))
            result = self.jobDB.setJobAttribute(job, "MinorStatus", minorStatus, update=True)
            if not result["OK"]:
                self.log.error(
                    "Failed setting MinorStatus", "%s for job %d: %s" % (minorStatus, job, result["Message"])
                )
                toRet = result

        if not minorStatus:  # Retain last minor status for stalled jobs
            result = self.jobDB.getJobAttributes(job, ["MinorStatus"])
            if result["OK"]:
                minorStatus = result["Value"]["MinorStatus"]
            else:
                self.log.error("Failed getting MinorStatus", "for job %d: %s" % (job, result["Message"]))
                minorStatus = "idem"
                toRet = result

        result = self.logDB.addLoggingRecord(job, status=status, minorStatus=minorStatus, source="StalledJobAgent")
        if not result["OK"]:
            self.log.warn("Failed adding logging record", result["Message"])
            toRet = result

        return toRet

    def _getProcessingType(self, jobID):
        """Get the Processing Type from the JDL, until it is promoted to a real Attribute"""
        processingType = "unknown"
        result = self.jobDB.getJobJDL(jobID, original=True)
        if not result["OK"]:
            return processingType
        classAdJob = ClassAd(result["Value"])
        if classAdJob.lookupAttribute("ProcessingType"):
            processingType = classAdJob.getAttributeString("ProcessingType")
        return processingType

    def _sendAccounting(self, jobID):
        """
        Send WMS accounting data for the given job.

        Run inside thread.
        """
        try:
            accountingReport = Job()
            endTime = "Unknown"
            lastHeartBeatTime = "Unknown"

            result = self.jobDB.getJobAttributes(jobID)
            if not result["OK"]:
                return result
            jobDict = result["Value"]

            startTime, endTime = self._checkLoggingInfo(jobID, jobDict)
            lastCPUTime, lastWallTime, lastHeartBeatTime = self._checkHeartBeat(jobID, jobDict)
            lastHeartBeatTime = fromString(lastHeartBeatTime)
            if lastHeartBeatTime is not None and lastHeartBeatTime > endTime:
                endTime = lastHeartBeatTime

            result = JobMonitoringClient().getJobParameter(jobID, "CPUNormalizationFactor")
            if not result["OK"] or not result["Value"]:
                self.log.error(
                    "Error getting Job Parameter CPUNormalizationFactor, setting 0",
                    result.get("Message", "No such value"),
                )
                cpuNormalization = 0.0
            else:
                cpuNormalization = float(result["Value"].get("CPUNormalizationFactor"))

        except Exception as e:
            self.log.exception(
                "Exception in _sendAccounting",
                "for job=%s: endTime=%s, lastHBTime=%s" % (str(jobID), str(endTime), str(lastHeartBeatTime)),
                lException=e,
            )
            return S_ERROR("Exception")
        processingType = self._getProcessingType(jobID)

        accountingReport.setStartTime(startTime)
        accountingReport.setEndTime(endTime)
        # execTime = toEpoch( endTime ) - toEpoch( startTime )
        # Fill the accounting data
        acData = {
            "Site": jobDict["Site"],
            "User": jobDict["Owner"],
            "UserGroup": jobDict["OwnerGroup"],
            "JobGroup": jobDict["JobGroup"],
            "JobType": jobDict["JobType"],
            "JobClass": jobDict["JobSplitType"],
            "ProcessingType": processingType,
            "FinalMajorStatus": JobStatus.FAILED,
            "FinalMinorStatus": JobMinorStatus.STALLED_PILOT_NOT_RUNNING,
            "CPUTime": lastCPUTime,
            "NormCPUTime": lastCPUTime * cpuNormalization,
            "ExecTime": lastWallTime,
            "InputDataSize": 0.0,
            "OutputDataSize": 0.0,
            "InputDataFiles": 0,
            "OutputDataFiles": 0,
            "DiskSpace": 0.0,
            "InputSandBoxSize": 0.0,
            "OutputSandBoxSize": 0.0,
            "ProcessedEvents": 0,
        }

        # For accidentally stopped jobs ExecTime can be not set
        if not acData["ExecTime"]:
            acData["ExecTime"] = acData["CPUTime"]
        elif acData["ExecTime"] < acData["CPUTime"]:
            acData["ExecTime"] = acData["CPUTime"]

        self.log.verbose("Accounting Report is:")
        self.log.verbose(acData)
        accountingReport.setValuesFromDict(acData)

        result = accountingReport.commit()
        if result["OK"]:
            self.jobDB.setJobAttribute(jobID, "AccountedFlag", "True")
        else:
            self.log.error("Failed to send accounting report", "Job: %d, Error: %s" % (int(jobID), result["Message"]))
        return result

    def _checkHeartBeat(self, jobID, jobDict):
        """Get info from HeartBeat"""
        result = self.jobDB.getHeartBeatData(jobID)
        lastCPUTime = 0
        lastWallTime = 0
        lastHeartBeatTime = jobDict["StartExecTime"]
        if lastHeartBeatTime == "None":
            lastHeartBeatTime = 0

        if result["OK"]:
            for name, value, heartBeatTime in result["Value"]:
                if name == "CPUConsumed":
                    try:
                        value = int(float(value))
                        if value > lastCPUTime:
                            lastCPUTime = value
                    except ValueError:
                        pass
                if name == "WallClockTime":
                    try:
                        value = int(float(value))
                        if value > lastWallTime:
                            lastWallTime = value
                    except ValueError:
                        pass
                if heartBeatTime > lastHeartBeatTime:
                    lastHeartBeatTime = heartBeatTime

        return lastCPUTime, lastWallTime, lastHeartBeatTime

    def _checkLoggingInfo(self, jobID, jobDict):
        """Get info from JobLogging"""
        logList = []
        result = self.logDB.getJobLoggingInfo(jobID)
        if result["OK"]:
            logList = result["Value"]

        startTime = jobDict["StartExecTime"]
        if not startTime or startTime == "None":
            # status, minor, app, stime, source
            for items in logList:
                if items[0] == "Running":
                    startTime = items[3]
                    break
            if not startTime or startTime == "None":
                startTime = jobDict["SubmissionTime"]

        if isinstance(startTime, six.string_types):
            startTime = fromString(startTime)
            if startTime is None:
                self.log.error("Wrong timestamp in DB", items[3])
                startTime = dateTime()

        endTime = dateTime()
        # status, minor, app, stime, source
        for items in logList:
            if items[0] == "Stalled":
                endTime = fromString(items[3])
        if endTime is None:
            self.log.error("Wrong timestamp in DB", items[3])
            endTime = dateTime()

        return startTime, endTime

    def _kickStuckJobs(self):
        """Reschedule jobs stuck in initialization status Rescheduled, Matched"""

        message = ""

        checkTime = dateTime() - self.matchedTime * second
        result = self.jobDB.selectJobs({"Status": JobStatus.MATCHED}, older=checkTime)
        if not result["OK"]:
            self.log.error("Failed to select jobs", result["Message"])
            return result

        jobIDs = result["Value"]
        if jobIDs:
            self.log.info("Rescheduling %d jobs stuck in Matched status" % len(jobIDs))
            result = self.jobDB.rescheduleJobs(jobIDs)
            if "FailedJobs" in result:
                message = "Failed to reschedule %d jobs stuck in Matched status" % len(result["FailedJobs"])

        checkTime = dateTime() - self.rescheduledTime * second
        result = self.jobDB.selectJobs({"Status": JobStatus.RESCHEDULED}, older=checkTime)
        if not result["OK"]:
            self.log.error("Failed to select jobs", result["Message"])
            return result

        jobIDs = result["Value"]
        if jobIDs:
            self.log.info("Rescheduling %d jobs stuck in Rescheduled status" % len(jobIDs))
            result = self.jobDB.rescheduleJobs(jobIDs)
            if "FailedJobs" in result:
                if message:
                    message += "\n"
                message += "Failed to reschedule %d jobs stuck in Rescheduled status" % len(result["FailedJobs"])

        if message:
            return S_ERROR(message)
        return S_OK()

    def _failSubmittingJobs(self):
        """Failed Jobs stuck in Submitting Status for a long time.
        They are due to a failed bulk submission transaction.
        """

        # Get old Submitting Jobs
        checkTime = dateTime() - self.submittingTime * second
        result = self.jobDB.selectJobs({"Status": JobStatus.SUBMITTING}, older=checkTime)
        if not result["OK"]:
            self.log.error("Failed to select jobs", result["Message"])
            return result

        for jobID in result["Value"]:
            result = self._updateJobStatus(jobID, JobStatus.FAILED)
            if not result["OK"]:
                self.log.error("Failed to update job status", result["Message"])
                continue

        return S_OK()

    def _sendKillCommand(self, job):
        """Send a kill signal to the job such that it cannot continue running.

        :param int job: ID of job to send kill command
        """
        ownerDN = self.jobDB.getJobAttribute(job, "OwnerDN")
        ownerGroup = self.jobDB.getJobAttribute(job, "OwnerGroup")
        if ownerDN["OK"] and ownerGroup["OK"]:
            wmsClient = WMSClient(
                useCertificates=True, delegatedDN=ownerDN["Value"], delegatedGroup=ownerGroup["Value"]
            )
            resKill = wmsClient.killJob(job)
            if not resKill["OK"]:
                self.log.error("Failed to send kill command to job", "%s: %s" % (job, resKill["Message"]))
        else:
            self.log.error(
                "Failed to get ownerDN or Group for job:",
                "%s: %s, %s" % (job, ownerDN.get("Message", ""), ownerGroup.get("Message", "")),
            )



"""The Request Task Agent takes request tasks created in the
TransformationDB and submits to the request management system.

The following options can be set for the RequestTaskAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN RequestTaskAgent
  :end-before: ##END
  :dedent: 2
  :caption: RequestTaskAgent options

* The options *SubmitTasks*, *MonitorTasks*, *MonitorFiles*, and *CheckReserved*
  need to be assigned any non-empty value to be activated

* .. versionadded:: v6r20p5

   It is possible to run the RequestTaskAgent without a *shifterProxy* or
   *ShifterCredentials*, in this case the credentials of the authors of the
   transformations are used to submit the jobs to the RMS. This enables the use of
   a single RequestTaskAgent for multiple VOs. See also the section about the
   :ref:`trans-multi-vo`.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.TransformationSystem.Agent.TaskManagerAgentBase import TaskManagerAgentBase

__RCSID__ = "$Id$"

AGENT_NAME = "Transformation/RequestTaskAgent"


class RequestTaskAgent(TaskManagerAgentBase):
    """An AgentModule to submit requests tasks"""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        TaskManagerAgentBase.__init__(self, *args, **kwargs)

        self.transType = []
        self.taskManager = None

    def initialize(self):
        """Standard initialize method"""
        res = TaskManagerAgentBase.initialize(self)
        if not res["OK"]:
            return res

        objLoader = ObjectLoader()
        _class = objLoader.loadObject("TransformationSystem.Client.RequestTasks", "RequestTasks")

        if not _class["OK"]:
            raise Exception(_class["Message"])

        self.requestTasksCls = _class["Value"]

        # clients
        self.taskManager = self.requestTasksCls(transClient=self.transClient)

        agentTSTypes = self.am_getOption("TransType", [])
        if agentTSTypes:
            self.transType = agentTSTypes
        else:
            self.transType = Operations().getValue("Transformations/DataManipulation", ["Replication", "Removal"])

        return S_OK()

    def _getClients(self, ownerDN=None, ownerGroup=None):
        """Set the clients for task submission.

        Here the taskManager becomes a RequestTasks object.

        See :func:`DIRAC.TransformationSystem.TaskManagerAgentBase._getClients`.
        """
        res = super(RequestTaskAgent, self)._getClients(ownerDN=ownerDN, ownerGroup=ownerGroup)
        threadTaskManager = self.requestTasksCls(ownerDN=ownerDN, ownerGroup=ownerGroup)
        res.update({"TaskManager": threadTaskManager})
        return res
